# SPDX-License-Identifier: Apache-2.0

import warnings
from logging import getLogger
import numpy as np
from onnx import GraphProto
from scipy.sparse import coo_matrix
from ..proto import TensorProto
from ..common.data_types import (
    _guess_type_proto_str, _guess_type_proto_str_inv)
from ..common._topology import (
    Variable, VariableStr, Scope, _update_domain_version, Operator,
    _get_main_opset_version, OPSET_TO_IR_VERSION)
from ..common._container import ModelComponentContainer
from ..common import utils
from ..common.data_types import guess_proto_type, _guess_numpy_type
from ..common._registration import _converter_pool, _shape_calculator_pool
from .._supported_operators import sklearn_operator_name_map
from ..proto import get_latest_tested_opset_version, onnx_proto
from ..proto.onnx_helper_modified import make_graph, make_model, from_array
from ..helpers.onnx_helper import infer_outputs
from .graph_state import GraphState, GraphStateVar
from .type_helper import _guess_type


logger = getLogger('skl2onnx')


class OnnxOperatorItem:
    """
    Accessor to one of the output returned by a *OnnxOperator*.

    :param onx_op: OnnxOperator
    :param index: integer
    """

    def __init__(self, onx_op, index, op_version=None):
        if not isinstance(index, int):
            raise TypeError("index must be an integer.")
        self.onx_op = onx_op
        self.index = index
        self.op_version = op_version

    def __str__(self):
        """
        usual
        """
        return "%s[%d]" % (str(self.onx_op), self.index)

    def get_latest_tested_opset_version(self):
        """
        Returns ``get_latest_tested_opset_version()``
        of the wrapped *OnnxOperator* instance.
        """
        return self.onx_op.get_latest_tested_opset_version()

    def add_to(self, scope, container, operator=None, run_converters=False):
        """
        Adds outputs to the container if not already added,
        registered the outputs if the node is not final.

        :param scope: scope
        :param container: container
        :param operator: overwrite inputs
        :param run_converters: must be True if called from method `to_onnx`
        """
        self.onx_op.add_to(scope, container, operator=operator,
                           run_converters=run_converters)

    def get_output_name(self, i=0):
        """
        Returns the output.
        """
        if i != 0:
            raise IndexError("Can only return the first item.")
        return self.onx_op.get_output_name(self.index)

    def get_output(self, i=0):
        """
        Returns the output.
        """
        if i != 0:
            raise IndexError("Can only return the first item.")
        return self.onx_op.get_output(self.index)

    @property
    def outputs(self):
        """
        Returns the outputs of the node.
        """
        if self.onx_op is None:
            raise RuntimeError(
                "self.onx_op cannot be None, type(self)={}".format(
                    type(self)))
        if self.index is None:
            raise RuntimeError(
                "self.index cannot be None, type(self)={}".format(
                    type(self)))
        outputs = self.onx_op.outputs
        if outputs is None:
            raise RuntimeError(
                "self.onx_op.outputs cannot be None, "
                "type(self)={}, type(self.onx_op)={}, "
                "type(self.onx_op.state)={}".format(
                    type(self), type(self.onx_op), type(self.onx_op.state)))
        return outputs[self.index:self.index + 1]

    def get_output_type_inference(self, input_shapes=None):
        """
        Returns the inferred shape.
        """
        if self.onx_op is None:
            raise RuntimeError(
                "self.onx_op cannot be None, type(self)={}".format(
                    type(self)))
        if self.index is None:
            raise RuntimeError(
                "self.index cannot be None, type(self)={}".format(
                    type(self)))
        outputs = self.onx_op.get_output_type_inference(input_shapes)
        if outputs is None:
            raise RuntimeError(
                "self.onx_op.outputs cannot be None, "
                "type(self)={}, type(self.onx_op)={}, "
                "type(self.onx_op.state)={}".format(
                    type(self), type(self.onx_op), type(self.onx_op.state)))
        return outputs[self.index:self.index + 1]


class OnnxOperator:
    """
    Ancestor to every *ONNX* operator exposed in
    :mod:`onnx_ops` and :mod:`onnx_ops_ml`. These files
    are automatically generated by unit test
    *test_onnx_operators_parse_spec*
    Every instance is supposed to be included in
    a graph as a node.

    :param inputs: list of inputs expected by the operator
    :param op_version: to select a specific version of the operator
    :param output_names: used defined names for the outputs
    :param domain: to overwrite the default domain
    :param global_context: operator *If* executes one subgraph
        whose nodes may use one existing output in the current
        context. If not used in the main graph, these operators
        are not linked to the output and cannot be retrieved.
        *global_context* is a dictionary mapped the subgraph input
        names to these operators.
    :param clear_subgraph_inputs: clears subgraphs outputs.
        Operator *If* does take subgraphs as attribute,
        there are subgraphs with no inputs and
        global variable as hidden inputs.
    :param kwargs: additional parameters of the operator

    .. versionchanged:: 1.10.1
        Parameter *global_context*, *clear_subgraph_inputs*
        were added.
    """
    class OnnxOperatorVariable(GraphStateVar):

        def __init__(self, index, name=None):
            self.index = index
            self.name = name

        def as_variable(self, scope):
            name = "ov%s" % self.name
            if (hasattr(self, "variable_") and
                    self.variable_.onnx_name == name):
                return self.variable_
            var = Variable(name, name, scope=scope, type=None)
            if scope is not None:
                scope.register_variable(var)
            self.variable_ = var
            return var

        def __repr__(self):
            return "OnnxOperatorVariable('%s')" % self.name

        def __iter__(self):
            yield self.name
            yield None

    class UnscopedVariable(GraphStateVar):
        def __init__(self, name):
            self.name = name

        def as_variable(self, scope):
            name = self.name
            if (hasattr(self, "variable_") and
                    self.variable_.onnx_name == name):
                return self.variable_
            if scope is not None:
                if name in scope.variables:
                    var = scope.variables[name]
                else:
                    onnx_name = scope.get_unique_variable_name(name)
                    var = Variable(name, onnx_name, scope=scope, type=None)
                    scope.register_variable(var)
                self.variable_ = var
            else:
                var = Variable(name, name, scope=scope, type=None)
            return var

        def __eq__(self, name):
            if isinstance(name, str):
                return name == self.name
            elif isinstance(name, OnnxOperator.UnscopedVariable):
                return self.name == name.name
            else:
                raise TypeError('Unsupported type for comparison {}'.format(
                    type(name)))

        def __repr__(self):
            return "UnscopedVariable('%s')" % self.name

        def __iter__(self):
            yield self.name
            yield None

    class ConstantVariable(GraphStateVar):
        def __init__(self, value):
            self.value = value

        def as_variable(self, scope):
            ha = utils.hash_array(self.value)
            name = "CST%s" % ha
            if (hasattr(self, "variable_") and
                    self.variable_.onnx_name == name):
                return self.variable_
            if scope is not None:
                var = scope.declare_local_variable(
                    name, type=_guess_type(self.value))
            else:
                var = Variable(name, name, scope=scope,
                               type=_guess_type(self.value))
            self.variable_ = var
            return var

        @property
        def ConstantValue(self):
            return self.value

        def __str__(self):
            return "Cst({})".format(self.value)

        def __iter__(self):
            yield "id%d" % id(self)
            yield _guess_type(self.value)

    def find_schema(self, op_version):
        """
        Checks if there is an existing schema for a
        specific version.

        :param op_version: requested version
        :return: schema
        """
        if not hasattr(self.__class__, 'past_version'):
            raise RuntimeError("Missing attribute 'past_version', there is "
                               "no other available schema.")
        found = None
        for v in self.past_version.values():
            if v.since_version > op_version:
                continue
            if found is None or v.since_version > found.since_version:
                found = v
        if found is None:
            raise RuntimeError(
                "Operator '{}': requested version {} < "
                "{} schema version.".format(
                    self.__class__.__name__,
                    op_version, self.since_version))
        return found

    def __init__(self, *inputs, op_version=None, output_names=None,
                 domain=None, global_context=None,
                 clear_subgraph_inputs=False, **kwargs):

        if (output_names is None and
                self.__class__.__name__.startswith("OnnxScan")):
            raise NotImplementedError(
                "The class cannot infer the number of variables "
                "for node '{}' yet. output_names must be specified"
                ".".format(self.__class__.__name__))
        if isinstance(output_names, (str, Variable)):
            output_names = [output_names]
            if isinstance(output_names[0], str):
                output_names[0] = VariableStr(output_names[0])
        elif isinstance(output_names, Operator):
            if len(output_names.outputs) == 0:
                raise ValueError(
                    "output_names cannot be empty (operator %r)."
                    "" % output_names)
            output_names = output_names.outputs.copy()
        elif isinstance(output_names, Operator.OperatorList):
            if len(output_names) == 0:
                raise ValueError(
                    "output_names cannot be empty (operator %r)."
                    "" % self.__class__.__name__)
            output_names = output_names.copy()
        elif isinstance(output_names, list):
            if len(output_names) == 0:
                raise ValueError(
                    "output_names cannot be empty (operator %r)."
                    "" % self.__class__.__name__)
            output_names = output_names.copy()
            for i in range(len(output_names)):
                if isinstance(output_names[i], str):
                    output_names[i] = VariableStr(output_names[i])
        elif output_names is not None:
            raise TypeError(
                "output_names must be a string or a list not %r."
                "" % type(output_names))

        if op_version is None:
            if domain == '':
                self.op_version = get_latest_tested_opset_version()
            else:
                self.op_version = None
        else:
            self.op_version = op_version
        self.since_version = self.__class__.since_version

        if (self.op_version is not None and
                self.op_version < self.since_version):
            schema = self.find_schema(self.op_version)
            self.since_version = schema.since_version
            self.expected_inputs = schema.expected_inputs.copy()
            self.expected_outputs = schema.expected_outputs.copy()
            self.input_range = schema.input_range
            self.output_range = schema.output_range
        else:
            self.expected_inputs = (
                None if self.__class__.expected_inputs is None
                else self.__class__.expected_inputs.copy())
            self.expected_outputs = (
                None if self.__class__.expected_outputs is None
                else self.__class__.expected_outputs.copy())
            self.input_range = self.__class__.input_range
            self.output_range = self.__class__.output_range
            if self.__class__.__name__ not in {
                    'OnnxScan', 'OnnxLoop', 'OnnxIf'}:
                # TODO: the minimum opset depends on embedded graph
                # by default, it takes the given op_version but the
                # optimal value could be lower.
                self.op_version = self.since_version
            if self.op_version is None:
                self.op_version = self.since_version

        if (self.op_version is not None and
                self.op_version < self.since_version):
            raise RuntimeError(
                "Operator '{}': requested version {} < "
                "{} schema version.".format(
                    self.__class__.__name__,
                    self.op_version, self.since_version))

        self.state = None
        self.domain = domain
        self.kwargs = kwargs
        self.onnx_prefix_name = None

        # check inputs
        if len(inputs) == 0:
            if self.input_range[0] == self.input_range[1]:
                self.inputs = [OnnxOperator.UnscopedVariable(_[0])
                               for _ in self.expected_inputs]
            else:
                # The number of inputs may vary.
                self.inputs = None
        else:
            self.inputs = []
            for inp in inputs:
                if isinstance(inp, str):
                    self.inputs.append(OnnxOperator.UnscopedVariable(inp))
                elif isinstance(inp, (OnnxOperator, Variable,
                                      OnnxOperatorItem, OnnxSubEstimator)):
                    self.inputs.append(inp)
                elif isinstance(inp, tuple) and len(inp) == 2:
                    self.inputs.append(inp)
                elif isinstance(inp, (np.ndarray, coo_matrix)):
                    self.inputs.append(
                        OnnxOperator.ConstantVariable(inp))
                elif isinstance(inp, TensorProto):
                    self.inputs.append(OnnxOperator.ConstantVariable(inp))
                elif isinstance(inp, (OnnxOperator.OnnxOperatorVariable,
                                      OnnxOperator.ConstantVariable)):
                    self.inputs.append(inp)
                elif isinstance(inp, (np.int64, np.float32,
                                      np.float64, np.bool_,
                                      np.int8, np.uint8)):
                    self.inputs.append(OnnxOperator.ConstantVariable(inp))
                elif isinstance(inp, (float, )):
                    self.inputs.append(np.float64(inp))
                elif isinstance(inp, (int, )):
                    self.inputs.append(np.int64(inp))
                else:
                    raise TypeError(
                        "Unable to interpret the input name for type {} in "
                        "operator '{}' (value={}).".format(
                            type(inp), self.__class__.__name__, inp))

        if self.inputs is not None:
            if (len(self.inputs) < self.input_range[0] or
                    len(self.inputs) > self.input_range[1]):
                raise RuntimeError(
                    "Operator '{}' expects a number of inputs "
                    "in [{}, {}] not {} (expected opset={}, "
                    "class opset={})".format(
                        self.operator_name, *self.input_range,
                        len(self.inputs), op_version, self.op_version))
        # global context
        if global_context is None:
            self.global_context = None
        else:
            if not isinstance(global_context, dict):
                raise TypeError(
                    "global_context must be a dictionary not %r."
                    "" % type(global_context))
            for k, v in global_context.items():
                if not isinstance(v, (OnnxOperator, OnnxOperatorItem)):
                    raise TypeError(
                        "Value %r in must be an OnnxOperator or an "
                        "OnnxOperatorItem not %r." % (k, type(v)))
            self.global_context = global_context

        # check output
        self.output_names = output_names
        self.output_variables = None

        if self.output_names is not None:
            if len(self.output_names) == 0:
                raise ValueError(
                    "output_names can be None but cannot be empty for "
                    "operator %r." % self)
            if self.output_variables is None:
                self.output_variables = [None for o in self.output_names]
            for i in range(len(self.output_names)):
                name = self.output_names[i]
                if isinstance(name, Variable):
                    self.output_variables[i] = name
                else:
                    raise TypeError("output_names must be a list of strings "
                                    "and element %r is %r (%r)" % (
                                        i, type(name), name))
            if all(map(lambda x: x is None, self.output_variables)):
                self.output_variables = None

        if (self.output_names is not None and (
                self.expected_outputs is None or
                len(self.output_names) > len(self.expected_outputs))):
            if self.expected_outputs is None:
                self.expected_outputs = []
            for i in range(len(self.expected_outputs),
                           len(self.output_names)):
                self.expected_outputs.append((self.output_names[i], None))

        if (self.expected_inputs is None or
                len(self.inputs) > len(self.expected_inputs)):
            if self.expected_inputs is None:
                self.expected_inputs = []
            for i in range(len(self.expected_inputs),
                           len(self.inputs)):
                inp = self.inputs[i]
                if isinstance(inp, GraphStateVar):
                    inp = tuple(inp)
                elif isinstance(inp, str):
                    inp = (inp, None)
                elif hasattr(inp, 'add_to'):
                    # OnnxOperator
                    existing = set(_[0] for _ in self.expected_inputs)
                    i = 10
                    name = "input%d" % (10 + i)
                    while name in existing:
                        i += 1
                        name = "input%d" % (10 + i)
                    inp = (name, None)
                self.expected_inputs.append(inp)

        self.output_names_ = None
        self._post_process_attributes(
            clear_subgraph_inputs=clear_subgraph_inputs)
        logger.debug(
            '[Ops] +%s-%d (%s) id=%d',
            self.__class__.__name__, self.op_version, self.domain, id(self))

    def _post_process_attributes(self, clear_subgraph_inputs=False):
        """
        Walks through attributes and replaces them by ONNX
        values.
        """
        # Looks into attributes if there is any tuple
        # (GraphProto, OnnxOperator). In that case, the function
        # replaces the tuple by the graph proto and keeps
        # in attributes graph_algebra the OnnxOperator
        # which is the source of it.
        updates = {}
        graph_algebra = {}
        for k, v in self.kwargs.items():
            if isinstance(v, tuple) and isinstance(v[0], GraphProto):
                updates[k] = v[0]
                graph_algebra[k] = v[1]
        if len(graph_algebra) > 0:
            self.kwargs.update(updates)
            self.graph_algebra = graph_algebra

        if clear_subgraph_inputs:
            for k, v in self.kwargs.items():
                if isinstance(v, GraphProto):
                    del v.input[:]

        if self.__class__.__name__ == "OnnxConstantOfShape":
            if "value" in self.kwargs:
                value = self.kwargs['value']
                if isinstance(value, TensorProto):
                    return
                if isinstance(value, np.ndarray):
                    if value.shape == (1, ):
                        val = value[0]
                    elif len(value.shape) == 0:
                        val = value
                    else:
                        raise RuntimeError(
                            "Unexpected shape %r for value, it must be "
                            "an array of one element." % value.shape)
                    self.kwargs['value'] = from_array(
                        np.array([val], dtype=value.dtype))
                    return
                raise TypeError(
                    "Unexpected type %r for value. It should be an array "
                    "of one element." % type(value))
            return

        if self.__class__.__name__ == "OnnxCast":
            if "to" in self.kwargs:
                value = self.kwargs['to']
                if isinstance(value, int):
                    return
                to = guess_proto_type(_guess_numpy_type(value, None))
                self.kwargs['to'] = to
            return

    def __str__(self):
        """
        usual
        """
        return "{}({} in) -> {}".format(
            self.__class__.__name__,
            len(self.inputs) if self.inputs is not None else 0,
            [str(o) for o in self.output_names]
            if self.output_names is not None else "?")

    def set_onnx_name_prefix(self, onnx_prefix_name):
        """
        Provides a name to define a prefix in the onnx graph
        to avoid to get unreadable node names. The method
        does not overwrite an existing name, it propagates
        the prefix to inputs and stops the propagation
        if the prefix is already defined.
        """
        if self.onnx_prefix_name is None:
            self.onnx_prefix_name = onnx_prefix_name
            for inp in self.inputs:
                if hasattr(inp, 'onnx_prefix_name'):
                    inp.set_onnx_name_prefix(onnx_prefix_name)
        return self

    @property
    def onnx_prefix(self):
        if self.onnx_prefix_name is None:
            name = self.__class__.__name__
            if name.startswith("Onnx"):
                name = name[4:]
            return name[:2]
        return self.onnx_prefix_name

    def __getitem__(self, index):
        """
        Returns an accessor to one of the output
        of this node.
        """
        return OnnxOperatorItem(self, index, self.op_version)

    def get_output_name(self, i, scope=None):
        "Returns name of output *i*."
        if self.state is not None:
            return self.state.computed_outputs_[i][0]
        if self.output_names_ is not None:
            return self.output_names_[i]
        self._set_output_names_(getattr(self, 'scope', None) or scope, None)
        return self.output_names_[i]

    def get_output(self, i, scope=None):
        "Returns name of output *i*."
        if self.state is not None:
            return self.state.computed_outputs_[i]
        if self.output_names_ is not None:
            res = self.output_names_[i]
            if not isinstance(res, (tuple, Variable)):
                raise RuntimeError(
                        "Unable to retrieve output %r from %r."
                        "" % (i, self))
            return res

    def _set_output_names_(self, scope, operator):
        "Called by add_to."
        if operator is not None:
            self.operator_ = operator
        if self.output_names_ is not None:
            raise RuntimeError(
                "output_names_ is already set.")
        elif self.output_variables is not None:
            outputs = [o.onnx_name for o in self.output_variables]
            self.output_names_ = outputs
        elif self.output_names:
            if not isinstance(self.output_names, (list, tuple)):
                louts = [self.output_names]
            else:
                louts = self.output_names
            if operator is not None and len(louts) != len(operator.outputs):
                raise RuntimeError(
                    "Output mismatch for '{}'\n{}\n{}".format(
                        type(operator.raw_operator),
                        louts, operator.outputs))
            outputs = []
            for iname, name in enumerate(louts):
                if name is None:
                    raise AssertionError(
                        "Issue for operator '{}'.".format(
                            type(operator.raw_operator)))
                if name.startswith('u(') and name[-1] == ')':
                    name = scope.get_unique_variable_name(name[2:-1])
                elif operator is not None:
                    oout = operator.outputs[iname]
                    name = oout.onnx_name
                outputs.append(name)
            self.output_names_ = outputs
        elif self.expected_outputs is None:
            raise AttributeError(
                "expected_outputs is None for operator=%r, output_names=%r, "
                "output_variables=%r, operator=%r" % (
                    self, self.output_names, self.output_variables, operator))
        else:
            if scope is None:
                raise RuntimeError("scope must not be None.")
            outputs = []
            for name in self.expected_outputs:
                name = scope.get_unique_variable_name(
                    self.onnx_prefix + "_" + name[0])
                outputs.append(name)
            self.output_names_ = outputs
        return outputs

    def _add_to_inputs(self, operator):
        inputs = []
        for input in self.inputs:
            if isinstance(input, OnnxOperator.OnnxOperatorVariable):
                if operator is None:
                    raise RuntimeError("A placeholder cannot be replaced "
                                       "as an operator is not specified.")
                if len(operator.inputs) == 0:
                    raise RuntimeError("No input variable in {}.".format(
                        operator))
                # The inputs must be looked into the graph.
                for i in operator.inputs:
                    if i.onnx_name == input.name:
                        inputs.append(i)
                        break
                else:
                    vars = ', '.join(map(lambda o: "'%s'" % o.onnx_name,
                                         operator.inputs))
                    raise RuntimeError("Unable to find variable "
                                       "{} in {}.".format(input, vars))
            else:
                inputs.append(input)
        return inputs

    def add_to(self, scope, container, operator=None, run_converters=False):
        """
        Adds outputs to the container if not already added,
        registered the outputs if the node is not final.

        :param scope: scope
        :param container: container
        :param operator: overwrite inputs
        :param run_converters: False by default, must be True if
            called from method `to_onnx`

        At this stage, inputs types are not necessarily known.
        """
        if self.state is None:
            if self.is_deprecated:
                raise RuntimeError(
                    "Node '{}' is deprecated. This API cannot deprecated "
                    "nodes.".format(self.__class__.__name__))
            if (self.op_version is not None and
                    self.op_version < self.since_version):
                raise RuntimeError(
                    "Incompatible versions for node '{}'  op_version {} "
                    "< since_version {}.".format(
                        self.__class__.__name__, self.op_version,
                        self.since_version))
            if self.kwargs.get('op_version', '') is None:
                kwargs = self.kwargs.copy()
                del kwargs['op_version']
            else:
                kwargs = self.kwargs

            self._set_output_names_(scope, operator)
            domain = self.domain
            if domain is None:
                domain = self.__class__.domain
            inputs = self._add_to_inputs(operator)

            logger.debug("[Ops.add_to] state id=%d", id(self))
            self.state = GraphState(
                inputs, self.output_names_, self.operator_name,
                scope, container, None, op_version=self.op_version,
                op_domain=domain, onnx_prefix_name=self.onnx_prefix,
                expected_inputs=self.expected_inputs,
                expected_outputs=self.expected_outputs,
                input_range=self.input_range,
                output_range=self.output_range,
                operator=operator, run_converters=run_converters, **kwargs)
            self.state.run()
        self._verify_add_to_()

    def _verify_add_to_(self):
        if self.state is None:
            raise RuntimeError(
                "Graph was not produced for operator '{}': {}."
                "".format(self.__class__.__name__, self))
        for i in self.inputs:
            if hasattr(i, '_verify_add_to_'):
                i._verify_add_to_()

    @property
    def outputs(self):
        """
        Returns the outputs of the node.
        """
        if self.state is None:
            raise RuntimeError("Method add_to was not called.")
        return self.state.outputs

    def get_output_type_inference(self, input_shapes=None):
        """
        Returns the expected output types in a list.
        """
        if (self.state is not None and
                self.state.computed_outputs_ is not None):
            return self.state.computed_outputs_

        expected_inputs = (
            self.state.computed_inputs_
            if self.expected_inputs is None else self.expected_inputs)
        if expected_inputs is None:
            raise RuntimeError(
                "Attribute 'expected_inputs' is empty for %r, "
                "input_shapes=%r." % (self, input_shapes))

        expected_outputs = (
            self.state.computed_outputs_
            if self.expected_outputs is None else self.expected_outputs)
        if expected_outputs is None:
            raise RuntimeError(
                "Attribute 'expected_outputs' is empty for %r, "
                "input_shapes=%r." % (self, input_shapes))

        # Shape inference only work on a full graph.
        if input_shapes is None:
            input_shapes = self.inputs

        given = {}
        for i, inp in enumerate(input_shapes):
            if isinstance(inp, tuple):
                given[i] = inp[1]
            elif isinstance(inp, GraphStateVar):
                dt = inp.as_variable(scope=None)
                if dt.type is None:
                    continue
                given[i] = dt.type
        rev = {}
        for i, (name, v) in enumerate(expected_inputs):
            if v in rev:
                rev[v].append(i)
            else:
                rev[v] = [i]

        res = []
        for name, ct in expected_outputs:
            if isinstance(ct, str) and ct[0] in ('T', 'V', 'I'):
                if (ct[0] not in rev or
                        all(map(lambda k: k not in given, rev[ct]))):
                    raise NotImplementedError(
                        "Unable to guess output type for (%r, %r) - "
                        "given=%r - rev=%r input_shapes=%r expected_inputs"
                        "=%r." % (
                            name, ct, given, rev, input_shapes,
                            self.expected_inputs))
                found = False
                for ind in rev[ct]:
                    if ind in given:
                        res.append((name, given[ind]))
                        found = True
                        break
                if not found:
                    raise NotImplementedError(
                        "Unable to guess output type for (%r, %r) - "
                        "given=%r - rev=%r input_shapes=%r expected_inputs"
                        "=%r." % (
                            name, ct, given, rev, input_shapes,
                            self.expected_inputs))
                continue
            if isinstance(ct, str):
                try:
                    dt = _guess_type_proto_str(ct, None)
                except NotImplementedError as e:
                    raise NotImplementedError(
                        "Unable to guess output type for (%r, %r) - "
                        "given=%r - rev=%r." % (name, ct, given, rev)) from e
                res.append((name, dt))
                continue
            try:
                dt = _guess_type_proto_str(
                    _guess_type_proto_str_inv(ct), None)
            except NotImplementedError as e:
                raise NotImplementedError(
                    "Unable to guess output type for (%r, %r) - given=%r - "
                    "rev=%r." % (name, ct, given, rev)) from e
            res.append((name, dt))

        return res

    def _clean_attributes(self, *args, recursive=True):
        """
        Removes attributes in this node and its parents.
        """
        for arg in args:
            if arg in ('state', 'output_names_'):
                setattr(self, arg, None)
            elif hasattr(self, arg):
                delattr(self, arg)
        if recursive:
            for obj in self.inputs:
                if isinstance(obj, OnnxOperator):
                    obj._clean_attributes(*args, recursive=True)

    def to_onnx(self, inputs=None, outputs=None, other_outputs=None,
                target_opset=None, domain=None, verbose=0):
        """
        Converts this operator into an ONNX graph.

        :param inputs: specific inputs (as a dictionary) or
            default inputs if not specified
        :param outputs: specific outputs
        :param other_outputs: additional outputs to consider
            as graph outputs but not outputs of this particular
            node
        :param target_opset: dictionary with target opset per domain,
            None for the default one
        :param domain: domain of the operator
        :param verbose: prints information
        """
        if isinstance(target_opset, dict):
            dom = self.domain or ''
            target_opset = target_opset.get(dom, None)
        elif isinstance(target_opset, int):
            if self.domain not in ('', None):
                # The target_opset is for the domain ''
                # We ignore it.
                target_opset = None
        elif target_opset is not None:
            raise TypeError(
                "target_opset must be a dictionary {domain: "
                "target_opset} not %r for operator %r." % (
                    target_opset, self.__class__.__name__))
        if self.domain in ('', None) and target_opset == 1:
            raise RuntimeError("target_opset cannot be 1.")
        if (self.op_version is not None and target_opset is not None and
                self.op_version > target_opset):
            raise RuntimeError(
                "target_opset={} is lower than the version={} requested "
                "for this node '{}'.".format(
                    target_opset, self.op_version, self.__class__.__name__))
        if self.state is not None:
            # The conversion already happened and needs to be cleaned.
            self._clean_attributes("output_names_", "state")
        if inputs is None:
            raise NotImplementedError("inputs must be specified.")
        if isinstance(inputs, dict):
            inputs = [(k, v) for k, v in inputs.items()]
        new_inputs = []
        for obj in inputs:
            if isinstance(obj, Variable):
                new_inputs.append((obj.onnx_name, obj.type))
            elif isinstance(obj, tuple) and len(obj) == 2:
                ty = _guess_type(obj[1])
                new_inputs.append((obj[0], ty))
            else:
                raise TypeError("Inputs must be Variable or "
                                "tuple(name, type) not {}."
                                "".format(type(obj)))
        inputs = new_inputs
        for name, typ in inputs:
            if typ is None:
                raise RuntimeError("Type input '{}' for operator '{}' "
                                   "is unknown. You should specify "
                                   "input types.".format(
                                       name, self.__class__.__name__))

        registered_models = dict(
            conv=_converter_pool, shape=_shape_calculator_pool,
            aliases=sklearn_operator_name_map)

        target_opset = self.get_latest_tested_opset_version(target_opset)
        container = ModelComponentContainer(
            target_opset, registered_models=registered_models)

        model_name = self.__class__.__name__
        logger.debug(
            "[Ops.to_onnx] %s id=%d",
            self.__class__.__name__, id(self))
        scope = Scope(model_name, target_opset=target_opset,
                      registered_models=registered_models)
        for inp in inputs:
            var = Variable(inp[0], inp[0], scope=scope, type=inp[1])
            container.add_input(var)
            scope.register_variable(var)
        self.add_to(scope, container, run_converters=True)

        extra_outputs = []
        if other_outputs is not None:
            extra_outputs.extend(other_outputs)
        if self.global_context is not None:
            for name, var in self.global_context.items():
                if var.output_names is None:
                    # The variable name is likely to be different.
                    from .onnx_ops import OnnxIdentity
                    var2 = OnnxIdentity(
                        var, op_version=var.op_version,
                        output_names=[name])
                else:
                    var2 = var
                extra_outputs.append(var2)
        for out in extra_outputs:
            if not hasattr(out, 'add_to'):
                raise RuntimeError(
                    "Extra outputs must have method 'add_to'.")
            out.add_to(scope, container, run_converters=True)
        logger.debug(
            "[Ops.to_onnx] %s id=%d extra_outputs=%r",
            self.__class__.__name__, id(self), extra_outputs)

        # infer shapes
        if outputs:
            if isinstance(outputs, dict):
                outputs = [(k, v) for k, v in outputs.items()]
            shapes = []
            for o in outputs:
                if isinstance(o, Variable):
                    shapes.append(o)
                elif isinstance(o, tuple):
                    if isinstance(o[1], np.ndarray):
                        type_shape = _guess_type(o[1])
                    else:
                        type_shape = o[1]
                    shapes.append(Variable(o[0], o[0], None, type_shape))
                else:
                    raise TypeError("Outputs must be Variable or "
                                    "tuple(name, type).")
            logger.debug(
                "[Ops.to_onnx] %s id=%d outputs=%r",
                self.__class__.__name__, id(self), outputs)
        else:
            if verbose > 0:
                print("[op.to_onnx] infer outputs")
            shapes = infer_outputs(container, container.inputs,
                                   initializer=container.initializers,
                                   target_opset=target_opset)
            if self.output_names:
                set_names = set(v.onnx_name if hasattr(v, 'onnx_name') else v
                                for v in self.output_names)
                shapes = [shape for shape in shapes
                          if shape.onnx_name in set_names]

        logger.debug(
            "[Ops.to_onnx] %s id=%d shapes=%r",
            self.__class__.__name__, id(self), shapes)
        if verbose > 0:
            print("[op.to_onnx] shapes=%r" % shapes)

        # add the output to the container
        for shape in shapes:
            container.add_output(shape)

        container.ensure_topological_order()
        if verbose >= 2:
            print("---NODES---")
            for node in container.nodes:
                print("  %s - %s: %r -> %r" % (
                    node.op_type, node.name, node.input, node.output))

        # convert the graph
        graph = make_graph(
            container.nodes, model_name, container.inputs,
            container.outputs, container.initializers)
        onnx_model = make_model(graph)

        # domains
        _update_domain_version(container, onnx_model)

        # metadata
        opv = min(target_opset,
                  _get_main_opset_version(onnx_model) or target_opset)
        irv = OPSET_TO_IR_VERSION.get(opv, onnx_proto.IR_VERSION)
        onnx_model.ir_version = irv
        onnx_model.producer_name = utils.get_producer()
        onnx_model.producer_version = utils.get_producer_version()
        onnx_model.domain = utils.get_domain()
        onnx_model.model_version = utils.get_model_version()
        return onnx_model

    def enumerate_nodes(self):
        """
        Iterates on all nodes of the graph.
        """
        yield self
        for input in self.inputs:
            if isinstance(input, OnnxOperator):
                for i in input.enumerate_nodes():
                    yield i

    def enumerate_variables(self):
        """
        Iterates on all nodes of the graph to find variables.
        Returns an iterator `(node, i)` which means
        `node.inputs[i]` is a variable.
        """
        for node in self.enumerate_nodes():
            if self.inputs:
                for i, input in enumerate(self.inputs):
                    if isinstance(input, (OnnxOperator.UnscopedVariable,
                                          Variable)):
                        yield (node, i)

    def enumerate_initial_types(self):
        """
        Retrieves iniatial types of the implemented functions.
        It goes through the graph and returns the name and types
        of all variables not computed by an intemediate node.

        :return: list of `(name, type)`
        """
        for node, i in self.enumerate_variables():
            input = node.inputs[i]
            if isinstance(input, Variable):
                yield (input.onnx_name, input.type)
            elif isinstance(input, OnnxOperator.UnscopedVariable):
                name = input.name
                typ = node.expected_inputs[i]
                yield (name, typ)

    def get_latest_tested_opset_version(self, target_opset=None):
        """
        Returns *op_version*, or the max of all results
        returned by these method applied on every input,
        or ``get_latest_tested_opset_version()``.
        """
        if target_opset is not None:
            return target_opset
        return get_latest_tested_opset_version()


class OnnxSubEstimator(OnnxOperator):
    """
    This operator is used to call the converter of a model
    while converting another one.
    See :ref:`l-custom-parser-alternative`.
    """

    since_version = 1
    expected_inputs = None
    expected_outputs = None
    input_range = [1, 1e9]
    output_range = [1, 1e9]

    def __init__(self, skl_op, *inputs, op_version=None,
                 output_names=None, domain=None, options=None,
                 input_types=None, **kwargs):
        OnnxOperator.__init__(
            self, *inputs, op_version=op_version,
            output_names=output_names, domain=domain, **kwargs)
        self.operator_instance = skl_op
        self.options = options
        if skl_op is None and input_types is not None:
            raise RuntimeError(
                "input_types is only used when a sub-operator is defined.")
        self.input_types = input_types

    def __repr__(self):
        return "%s(%r, %s, op_version=%r, output_names=%r)" % (
            self.__class__.__name__, self.operator_instance,
            ", ".join("%r" % i for i in self.inputs),
            self.op_version, self.output_names)

    def add_to(self, scope, container, operator=None, run_converters=False):
        """
        Adds outputs to the container if not already added,
        registered the outputs if the node is not final.

        :param scope: scope
        :param container: container
        :param operator: overwrite inputs
        :param run_converters: must be True if called from method `to_onnx`
        """
        if self.state is None:
            if self.kwargs.get('op_version', '') is None:
                kwargs = self.kwargs.copy()
                del kwargs['op_version']
            else:
                kwargs = self.kwargs

            if self.output_names_ is not None:
                pass
            elif operator is not None:
                self.output_names_ = operator.outputs
            elif self.output_names:
                if not isinstance(self.output_names, (list, tuple)):
                    louts = [self.output_names]
                else:
                    louts = self.output_names
                outputs = []
                for name in louts:
                    if (isinstance(name, str) and name.startswith('u(') and
                            name[-1] == ')'):
                        name = VariableStr(
                            scope.get_unique_variable_name(name[2:-1]),
                            scope=scope)
                    if (isinstance(name, Variable) and
                            name.raw_name.startswith('u(') and
                            name.raw_name[-1] == ')'):
                        name = VariableStr(
                            scope.get_unique_variable_name(
                                name.raw_name[2:-1]),
                            scope=scope, type=name.type)
                    outputs.append(name)
                self.output_names_ = outputs
            else:
                self.output_names_ = None

            inputs = []
            for input in self.inputs:
                if isinstance(input, OnnxOperator.OnnxOperatorVariable):
                    if operator is None:
                        raise RuntimeError("A placeholder cannot be replaced "
                                           "as an operator is not specified.")
                    if len(operator.inputs) == 0:
                        raise RuntimeError("No input variable in {}.".format(
                            operator))
                    # The inputs must be looked into the graph.
                    for i in operator.inputs:
                        if i.onnx_name == input.name:
                            inputs.append(i)
                            break
                    else:
                        vars = ', '.join(map(lambda o: "'%s'" % o.onnx_name,
                                             operator.inputs))
                        raise RuntimeError("Unable to find variable "
                                           "{} in {}.".format(input, vars))
                elif isinstance(input, tuple) and len(input) == 2:
                    if scope is not None and input[0] in scope.variables:
                        var = scope.variables[input[0]]
                    else:
                        var = Variable(input[0], input[0], scope=scope,
                                       type=input[1])
                        if scope is not None:
                            scope.register_variable(var)
                    inputs.append(var)
                else:
                    inputs.append(input)

            logger.debug("[SubOps.add_to] state id=%d", id(self))
            self.state = GraphState(
                inputs, self.output_names_, self.operator_instance,
                scope, container, None, op_version=self.op_version,
                op_domain=None, onnx_prefix_name=self.onnx_prefix,
                options=self.options, run_converters=run_converters,
                input_types=self.input_types, **kwargs)
            self.state.run()


class WrappedModelAlias:

    def __init__(self, model, alias):
        self.model = model
        self.alias = alias


class OnnxSubOperator(OnnxSubEstimator):
    """
    This class is deprecated and will be removed in version 1.9.
    It should be replaced by :class:`OnnxSubEstimator
    <skl2onnx.algebra.onnx_operator.OnnxSubEstimator>`.
    """

    def __init__(self, *args, **kwargs):
        OnnxSubEstimator.__init__(self, *args, **kwargs)
        warnings.warn(("Class OnnxSubOperator will be removed in 1.10. "
                       "It should be replaced by OnnxSubEstimator."),
                      DeprecationWarning)



# SPDX-License-Identifier: Apache-2.0

from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_slice, apply_cast, apply_identity, apply_reshape)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def _common_convert_sklearn_zipmap(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    zipmap_attrs = {'name': scope.get_unique_operator_name('ZipMap')}
    to_type = onnx_proto.TensorProto.INT64

    if hasattr(operator, 'classlabels_int64s'):
        zipmap_attrs['classlabels_int64s'] = operator.classlabels_int64s
    elif hasattr(operator, 'classlabels_strings'):
        zipmap_attrs['classlabels_strings'] = operator.classlabels_strings
        to_type = onnx_proto.TensorProto.STRING

    if to_type == onnx_proto.TensorProto.STRING:
        apply_identity(scope, operator.inputs[0].full_name,
                       operator.outputs[0].full_name, container)
    else:
        apply_cast(scope, operator.inputs[0].full_name,
                   operator.outputs[0].full_name, container, to=to_type)
    return zipmap_attrs


def convert_sklearn_zipmap(scope: Scope, operator: Operator,
                           container: ModelComponentContainer):
    if len(operator.inputs) == 2:
        zipmap_attrs = _common_convert_sklearn_zipmap(
            scope, operator, container)
        container.add_node('ZipMap', operator.inputs[1].full_name,
                           operator.outputs[1].full_name,
                           op_domain='ai.onnx.ml', **zipmap_attrs)
        return

    if hasattr(operator, 'classlabels_int64s'):
        zipmap_attrs = dict(classlabels_int64s=operator.classlabels_int64s)
    elif hasattr(operator, 'classlabels_strings'):
        zipmap_attrs = dict(classlabels_strings=operator.classlabels_strings)
    else:
        raise RuntimeError(
            "operator should have attribute 'classlabels_int64s' or "
            "'classlabels_strings'.")
    container.add_node('ZipMap', operator.inputs[0].full_name,
                       operator.outputs[0].full_name,
                       op_domain='ai.onnx.ml', **zipmap_attrs)


def convert_sklearn_zipmap_columns(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    _common_convert_sklearn_zipmap(scope, operator, container)
    probs = operator.inputs[1].full_name
    for i in range(1, len(operator.outputs)):
        out = operator.outputs[i].full_name
        flat = scope.get_unique_variable_name(out)
        apply_slice(
            scope, probs, flat, container, starts=[i-1], ends=[i], axes=[1],
            operator_name=scope.get_unique_operator_name('Slice'))
        apply_reshape(
            scope, flat, out, container, desired_shape=(-1, ),
            operator_name=scope.get_unique_operator_name('reshape'))


register_converter('SklearnZipMap', convert_sklearn_zipmap)
register_converter('SklearnZipMapColumns', convert_sklearn_zipmap_columns)



# SPDX-License-Identifier: Apache-2.0

import warnings
import numpy as np
from onnx import TensorProto
from ..common._registration import register_converter
from ..common.data_types import (
    BooleanTensorType, Int64TensorType,
    guess_numpy_type, guess_proto_type)
from ..algebra.onnx_ops import (
    OnnxCast, OnnxLess, OnnxMul, OnnxAdd, OnnxDiv,
    OnnxGather, OnnxReduceMean, OnnxMax, OnnxSqueezeApi11)
from .nearest_neighbours import onnx_nearest_neighbors_indices_k


def convert_sklearn_local_outlier_factor(
        scope, operator, container, op_type='TreeEnsembleRegressor',
        op_domain='ai.onnx.ml', op_version=1):
    op = operator.raw_operator
    if not op.novelty:
        raise RuntimeError(
            "The converter only converts the model %r is novelty is True."
            "" % type(op))
    outputs = operator.outputs
    opv = container.target_opset
    options = container.get_options(
        op, dict(score_samples=None, optim=None))

    X = operator.inputs[0]
    dtype = guess_numpy_type(operator.inputs[0].type)
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        X = OnnxCast(X, to=proto_dtype, op_version=opv)

    metric = (op.effective_metric_ if hasattr(op, 'effective_metric_') else
              op.metric)
    neighb = op._fit_X.astype(dtype)
    k = op.n_neighbors_
    kwargs = {}
    if op.p != 2:
        if options['optim'] == 'cdist':
            warnings.warn(
                "Option p=%r may not be compatible with the runtime. "
                "See https://github.com/microsoft/onnxruntime/blob/master/"
                "docs/ContribOperators.md#com.microsoft.CDist.")
        kwargs['p'] = op.p

    top_k, dist = onnx_nearest_neighbors_indices_k(
        X, neighb, k, metric, dtype=dtype,
        op_version=opv, keep_distances=True,
        optim=options.get('optim', None),
        **kwargs)

    # dist_k = self._distances_fit_X_[neighbors_indices, self.n_neighbors_ - 1]
    # reach_dist_array = np.maximum(distances_X, dist_k)
    dist_k_ = OnnxGather(op._distances_fit_X_.astype(dtype),
                         top_k, op_version=opv)
    dist_k = OnnxSqueezeApi11(
        OnnxGather(dist_k_, np.array([op.n_neighbors_ - 1],
                                     dtype=np.int64),
                   axis=2, op_version=opv),
        axes=[2], op_version=opv)
    dist_k.set_onnx_name_prefix('dist_k')
    reach_dist_array = OnnxMax(
        OnnxMul(dist, np.array([-1], dtype=dtype), op_version=opv),
        dist_k, op_version=opv)

    # X_lrd=  return 1.0 / (np.mean(reach_dist_array, axis=1) + 1e-10)
    X_lrd = OnnxDiv(
        np.array([1], dtype=dtype),
        OnnxAdd(
            OnnxReduceMean(reach_dist_array, axes=[1],
                           op_version=opv, keepdims=1),
            np.array([1e-10], dtype=dtype), op_version=opv),
        op_version=opv)
    X_lrd.set_onnx_name_prefix('X_lrd')

    # lrd_ratios_array = self._lrd[neighbors_indices_X] / X_lrd[:, np.newaxis]
    lrd_ratios_array = OnnxDiv(
        OnnxGather(op._lrd.astype(dtype), top_k, op_version=opv),
        X_lrd, op_version=opv)
    lrd_ratios_array.set_onnx_name_prefix('lrd_ratios_array')

    # -np.mean(lrd_ratios_array, axis=1)
    if options['score_samples']:
        output_names_score_samples = [outputs[2]]
    else:
        output_names_score_samples = None
    score_samples = OnnxReduceMean(
        lrd_ratios_array, axes=[1], op_version=opv)
    score_samples.set_onnx_name_prefix('score_samples')
    score_samples_neg = OnnxMul(
        score_samples, np.array([-1], dtype=dtype), op_version=opv,
        output_names=output_names_score_samples)
    final = OnnxAdd(score_samples_neg, np.array([-op.offset_], dtype=dtype),
                    op_version=opv, output_names=[outputs[1]])

    # labels
    # is_inlier = np.ones(X.shape[0], dtype=int)
    # is_inlier[self.decision_function(X) < 0] = -1

    predict = OnnxAdd(
        OnnxMul(
            OnnxCast(
                OnnxLess(final, np.array([0], dtype=dtype), op_version=opv),
                to=TensorProto.INT64, op_version=opv),
            np.array([-2], dtype=np.int64), op_version=opv),
        np.array([1], dtype=np.int64), op_version=opv,
        output_names=outputs[0].full_name)
    predict.set_onnx_name_prefix('predict')

    predict.add_to(scope, container)
    final.add_to(scope, container)
    if options['score_samples']:
        score_samples_neg.add_to(scope, container)


register_converter('SklearnLocalOutlierFactor',
                   convert_sklearn_local_outlier_factor,
                   options={'score_samples': [True, False],
                            'optim': [None, 'cdist']})



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatType, Int64Type, StringType, TensorType,
    DoubleType, BooleanTensorType, FloatTensorType,
    Int64TensorType, StringTensorType,
    DoubleTensorType)
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_concat(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    for i in range(len(operator.inputs)):
        if len(operator.inputs[i].type.shape) != 2:
            operator.outputs[0].type.shape = [None, None]
            return

    N = operator.inputs[0].get_first_dimension()
    C = 0
    seen_types = []
    for i in operator.inputs:
        if C is not None:
            if isinstance(i.type, TensorType):
                if i.type.shape[1] is None:
                    C = None
                else:
                    C += i.type.shape[1]
            elif isinstance(i.type, (
                    Int64Type, FloatType, StringType, DoubleType)):
                C += 1
            else:
                C = None
        if len(seen_types) == 0:
            seen_types.append(i.type)

    def more_generic(t1, t2):
        if isinstance(t1, TensorType):
            if not isinstance(t2, TensorType):
                raise RuntimeError(
                    "Cannot merge columns with types {} and {}."
                    "Inputs:\n{}\nOutputs:\n{}".format(
                        t1, t2, operator.inputs, operator.outputs))
            for ts in [StringTensorType, DoubleTensorType, FloatTensorType,
                       Int64TensorType, BooleanTensorType]:
                if isinstance(t1, ts) or isinstance(t2, ts):
                    return ts
            raise RuntimeError(
                "Cannot merge columns with types {} and {}."
                "Inputs:\n{}\nOutputs:\n{}".format(
                    t1, t2, operator.inputs, operator.outputs))
        raise NotImplementedError(
            "Columns must be tensors."
            "Inputs:\n{}\nOutputs:\n{}".format(
                operator.inputs, operator.outputs))

    # Let's determine the resulting type
    final_type = None
    for seen in seen_types:
        if final_type is None:
            final_type = seen
        elif seen != final_type:
            merged_type = more_generic(final_type, seen)
            if merged_type:
                final_type = seen

    if final_type is None:
        raise NotImplementedError(
            "Columns must be tensors.\n"
            "- Inputs: {}\n- Outputs: {}\n- types: {}"
            "".format(
                operator.inputs, operator.outputs, seen_types))
    if final_type != operator.outputs[0].type:
        operator.outputs[0].type = type(final_type)([N, C])
    else:
        operator.outputs[0].type.shape = [N, C]


register_shape_calculator('SklearnConcat', calculate_sklearn_concat)
register_shape_calculator('SklearnGenericUnivariateSelect',
                          calculate_sklearn_concat)
register_shape_calculator('SklearnMultiply', calculate_sklearn_concat)
register_shape_calculator('SklearnRFE', calculate_sklearn_concat)
register_shape_calculator('SklearnRFECV', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectFdr', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectFpr', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectFromModel', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectFwe', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectKBest', calculate_sklearn_concat)
register_shape_calculator('SklearnSelectPercentile', calculate_sklearn_concat)
register_shape_calculator('SklearnVarianceThreshold', calculate_sklearn_concat)



# SPDX-License-Identifier: Apache-2.0

"""
Common errors.
"""


_missing_converter = """
It usually means the pipeline being converted contains a
transformer or a predictor with no corresponding converter
implemented in sklearn-onnx. If the converted is implemented
in another library, you need to register
the converted so that it can be used by sklearn-onnx (function
update_registered_converter). If the model is not yet covered
by sklearn-onnx, you may raise an issue to
https://github.com/onnx/sklearn-onnx/issues
to get the converter implemented or even contribute to the
project. If the model is a custom model, a new converter must
be implemented. Examples can be found in the gallery.
"""


class MissingShapeCalculator(RuntimeError):
    """
    Raised when there is no registered shape calculator
    for a machine learning operator.
    """

    def __init__(self, msg):
        super().__init__(msg + _missing_converter)


class MissingConverter(RuntimeError):
    """
    Raised when there is no registered converter
    for a machine learning operator. If the model is
    part of scikit-learn, you may raise an issue at
    https://github.com/onnx/sklearn-onnx/issues.
    """

    def __init__(self, msg):
        super().__init__(msg + _missing_converter)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import (
    apply_cast, apply_sub)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_proto_type)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_sgd_oneclass_svm(scope: Scope, operator: Operator,
                                     container: ModelComponentContainer):

    input_name = operator.inputs[0].full_name
    output_names = operator.output_full_names
    model = operator.raw_operator
    coef = model.coef_.T
    offset = model.offset_

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if isinstance(operator.inputs[0].type,
                  (BooleanTensorType, Int64TensorType)):
        cast_input_name = scope.get_unique_variable_name('cast_input')
        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=proto_dtype)
        input_name = cast_input_name

    coef_name = scope.get_unique_variable_name('coef')
    container.add_initializer(coef_name, proto_dtype,
                              coef.shape, coef.ravel())

    offset_name = scope.get_unique_variable_name('offset')
    container.add_initializer(offset_name, proto_dtype, offset.shape, offset)

    matmul_result_name = scope.get_unique_variable_name('matmul_result')
    container.add_node('MatMul', [input_name, coef_name], matmul_result_name,
                       name=scope.get_unique_operator_name('MatMul'))

    apply_sub(scope, [matmul_result_name, offset_name], output_names[1],
              container, broadcast=0)

    pred = scope.get_unique_variable_name('class_prediction')
    container.add_node('Sign', output_names[1], pred, op_version=9)
    apply_cast(scope, pred, output_names[0],
               container, to=onnx_proto.TensorProto.INT64)


register_converter('SklearnSGDOneClassSVM', convert_sklearn_sgd_oneclass_svm)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from sklearn.linear_model import (
    LogisticRegression,
    RidgeClassifier,
    RidgeClassifierCV,
)
from sklearn.svm import LinearSVC
from ..common._apply_operation import (
    apply_cast, apply_add, apply_sigmoid, apply_softmax,
    apply_normalizer)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    BooleanTensorType, DoubleTensorType, guess_proto_type)
from ..common.utils_classifier import (
    get_label_classes, _finalize_converter_classes)
from ..proto import onnx_proto


def apply_logistic(scope, input_name, output_name, container,
                   proto_dtype):
    sig_name = scope.get_unique_variable_name(input_name + "sig")
    apply_sigmoid(scope, input_name, sig_name, container)
    apply_normalizer(
        scope, sig_name, output_name, container, norm='L1',
        use_float=proto_dtype == onnx_proto.TensorProto.FLOAT)


def convert_sklearn_linear_classifier(scope: Scope, operator: Operator,
                                      container: ModelComponentContainer):
    op = operator.raw_operator
    coefficients = op.coef_.flatten().astype(float).tolist()
    classes = get_label_classes(scope, op)
    number_of_classes = len(classes)
    use_linear_op = container.is_allowed({'LinearClassifier'})

    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']

    if isinstance(op.intercept_, (float, np.float32)) and op.intercept_ == 0:
        # fit_intercept = False
        intercepts = ([0.0] * number_of_classes if number_of_classes != 2 else
                      [0.0])
    else:
        intercepts = op.intercept_.tolist()

    if number_of_classes == 2:
        coefficients = list(map(lambda x: -1 * x, coefficients)) + coefficients
        intercepts = list(map(lambda x: -1 * x, intercepts)) + intercepts

    multi_class = 0
    if hasattr(op, 'multi_class'):
        if op.multi_class == 'ovr':
            multi_class = 1
        else:
            multi_class = 2

    classifier_type = 'LinearClassifier'
    classifier_attrs = {
        'name': scope.get_unique_operator_name(classifier_type)
    }

    classifier_attrs['coefficients'] = coefficients
    classifier_attrs['intercepts'] = intercepts
    classifier_attrs['multi_class'] = 1 if multi_class == 2 else 0
    if (use_raw_scores or
            isinstance(op, (LinearSVC, RidgeClassifier, RidgeClassifierCV))):
        classifier_attrs['post_transform'] = 'NONE'
    elif isinstance(op, LogisticRegression):
        ovr = (op.multi_class in ["ovr", "warn"] or
               (op.multi_class == 'auto' and (op.classes_.size <= 2 or
                                              op.solver == 'liblinear')))
        classifier_attrs['post_transform'] = (
            'LOGISTIC' if ovr else 'SOFTMAX')
    else:
        classifier_attrs['post_transform'] = (
            'LOGISTIC' if multi_class > 2 else 'SOFTMAX')

    if all(isinstance(i, str) for i in classes):
        class_labels = [str(i) for i in classes]
        classifier_attrs['classlabels_strings'] = class_labels
    elif all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
        class_labels = [int(i) for i in classes]
        classifier_attrs['classlabels_ints'] = class_labels
    else:
        raise RuntimeError('Label vector must be a string or a integer '
                           'tensor.')

    if (not use_linear_op or
            type(operator.inputs[0].type) in (DoubleTensorType, )):
        # Double -> double parameters not supported in ONNX LinearClassifier
        proto_dtype = guess_proto_type(operator.inputs[0].type)
        coef = scope.get_unique_variable_name('coef')
        model_coef = np.array(
            classifier_attrs['coefficients'], dtype=np.float64)
        model_coef = model_coef.reshape((number_of_classes, -1)).T
        container.add_initializer(
            coef, proto_dtype, model_coef.shape, model_coef.ravel().tolist())
        intercept = scope.get_unique_variable_name('intercept')
        model_intercept = np.array(
            classifier_attrs['intercepts'], dtype=np.float64)
        model_intercept = model_intercept.reshape((number_of_classes, -1)).T
        container.add_initializer(
            intercept, proto_dtype, model_intercept.shape,
            model_intercept.ravel().tolist())
        multiplied = scope.get_unique_variable_name('multiplied')
        container.add_node(
            'MatMul', [operator.inputs[0].full_name, coef], multiplied,
            name=scope.get_unique_operator_name('MatMul'))

        if use_raw_scores:
            raw_score_name = operator.outputs[1].full_name
        else:
            raw_score_name = scope.get_unique_variable_name('raw_scores')
        apply_add(scope, [multiplied, intercept], raw_score_name, container)

        argmax_output_name = scope.get_unique_variable_name('label')
        container.add_node('ArgMax', raw_score_name, argmax_output_name,
                           name=scope.get_unique_operator_name('ArgMax'),
                           axis=1)
        _finalize_converter_classes(
            scope, argmax_output_name, operator.outputs[0].full_name,
            container, np.array(class_labels),
            onnx_proto.TensorProto.DOUBLE)

        if use_raw_scores:
            return

        if classifier_attrs['post_transform'] == 'LOGISTIC':
            apply_logistic(scope, raw_score_name,
                           operator.outputs[1].full_name, container,
                           proto_dtype=onnx_proto.TensorProto.DOUBLE)
            return
        elif classifier_attrs['post_transform'] == 'SOFTMAX':
            apply_softmax(scope, raw_score_name,
                          operator.outputs[1].full_name, container)
            return

        raise NotImplementedError(
            "post_transform '{}' is not supported with double.".format(
                classifier_attrs['post_transform']))

    label_name = operator.outputs[0].full_name
    input_name = operator.inputs[0].full_name
    if type(operator.inputs[0].type) == BooleanTensorType:
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, input_name, cast_input_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        input_name = cast_input_name

    if use_raw_scores:
        container.add_node(classifier_type, input_name,
                           [label_name, operator.outputs[1].full_name],
                           op_domain='ai.onnx.ml', **classifier_attrs)
    elif (isinstance(op, (LinearSVC, RidgeClassifier, RidgeClassifierCV))
            and op.classes_.shape[0] <= 2):
        raw_scores_tensor_name = scope.get_unique_variable_name(
            'raw_scores_tensor')
        positive_class_index_name = scope.get_unique_variable_name(
            'positive_class_index')

        container.add_initializer(positive_class_index_name,
                                  onnx_proto.TensorProto.INT64, [], [1])

        if (hasattr(op, '_label_binarizer') and
                op._label_binarizer.y_type_ == 'multilabel-indicator'):
            y_pred_name = scope.get_unique_variable_name('y_pred')
            binarised_label_name = scope.get_unique_variable_name(
                'binarised_label')

            container.add_node(classifier_type, input_name,
                               [y_pred_name, raw_scores_tensor_name],
                               op_domain='ai.onnx.ml', **classifier_attrs)
            container.add_node(
                'Binarizer', raw_scores_tensor_name, binarised_label_name,
                op_domain='ai.onnx.ml')
            apply_cast(
                scope, binarised_label_name, label_name,
                container, to=onnx_proto.TensorProto.INT64)
        else:
            container.add_node(classifier_type, input_name,
                               [label_name, raw_scores_tensor_name],
                               op_domain='ai.onnx.ml', **classifier_attrs)
        container.add_node(
            'ArrayFeatureExtractor',
            [raw_scores_tensor_name, positive_class_index_name],
            operator.outputs[1].full_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    else:
        # Make sure the probability sum is 1 over all classes
        if multi_class > 0 and not isinstance(
                op, (LinearSVC, RidgeClassifier, RidgeClassifierCV)):
            probability_tensor_name = scope.get_unique_variable_name(
                'probability_tensor')
            container.add_node(classifier_type, input_name,
                               [label_name, probability_tensor_name],
                               op_domain='ai.onnx.ml', **classifier_attrs)
            use_float = type(operator.inputs[0].type) not in (
                DoubleTensorType, )
            apply_normalizer(
                scope, probability_tensor_name, operator.outputs[1].full_name,
                container, norm='L1', use_float=use_float)
        elif (hasattr(op, '_label_binarizer') and
              op._label_binarizer.y_type_ == 'multilabel-indicator'):
            y_pred_name = scope.get_unique_variable_name('y_pred')
            binarised_label_name = scope.get_unique_variable_name(
                'binarised_label')

            container.add_node(
                classifier_type, input_name,
                [y_pred_name, operator.outputs[1].full_name],
                op_domain='ai.onnx.ml', **classifier_attrs)
            container.add_node(
                'Binarizer', operator.outputs[1].full_name,
                binarised_label_name, op_domain='ai.onnx.ml')
            apply_cast(
                scope, binarised_label_name, label_name,
                container, to=onnx_proto.TensorProto.INT64)
        else:
            container.add_node(classifier_type, input_name,
                               [label_name, operator.outputs[1].full_name],
                               op_domain='ai.onnx.ml', **classifier_attrs)


register_converter('SklearnLinearClassifier',
                   convert_sklearn_linear_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})
register_converter('SklearnLinearSVC', convert_sklearn_linear_classifier,
                   options={'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_array_feature_extractor(
        scope: Scope, operator: Operator, container: ModelComponentContainer):
    """
    Extracts a subset of columns. This is used by *ColumnTransformer*.
    """
    column_indices_name = scope.get_unique_variable_name('column_indices')

    for i, ind in enumerate(operator.column_indices):
        if not isinstance(ind, int):
            raise RuntimeError(("Column {0}:'{1}' indices must be specified "
                                "as integers. This error may happen when "
                                "column names are used to define a "
                                "ColumnTransformer. Column name in input data "
                                "do not necessarily match input variables "
                                "defined for the ONNX model.").format(i, ind))
    container.add_initializer(column_indices_name,
                              onnx_proto.TensorProto.INT64,
                              [len(operator.column_indices)],
                              operator.column_indices)

    container.add_node(
        'ArrayFeatureExtractor',
        [operator.inputs[0].full_name, column_indices_name],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
        op_domain='ai.onnx.ml')


register_converter('SklearnArrayFeatureExtractor',
                   convert_sklearn_array_feature_extractor)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_mul
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_multiply(scope: Scope, operator: Operator,
                             container: ModelComponentContainer):
    operand_name = scope.get_unique_variable_name(
        'operand')

    container.add_initializer(operand_name, onnx_proto.TensorProto.FLOAT,
                              [], [operator.operand])

    apply_mul(scope, [operator.inputs[0].full_name, operand_name],
              operator.outputs[0].full_name, container)


register_converter('SklearnMultiply', convert_sklearn_multiply)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.data_types import _guess_numpy_type
from ..common.shape_calculator import calculate_linear_regressor_output_shapes


def calculate_sklearn_cast(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=1)


def calculate_sklearn_cast_transformer(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=1)
    op = operator.raw_operator
    otype = _guess_numpy_type(op.dtype, operator.inputs[0].type.shape)
    operator.outputs[0].type = otype


register_shape_calculator(
    'SklearnCast', calculate_sklearn_cast)
register_shape_calculator(
    'SklearnCastTransformer', calculate_sklearn_cast_transformer)
register_shape_calculator('SklearnCastRegressor',
                          calculate_linear_regressor_output_shapes)



# SPDX-License-Identifier: Apache-2.0


from .._supported_operators import sklearn_operator_name_map
from ..common._apply_operation import apply_identity
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_ransac_regressor(scope: Scope, operator: Operator,
                                     container: ModelComponentContainer):
    """
    Converter for RANSACRegressor.
    """
    ransac_op = operator.raw_operator
    op_type = sklearn_operator_name_map[type(ransac_op.estimator_)]
    this_operator = scope.declare_local_operator(op_type, ransac_op.estimator_)
    this_operator.inputs = operator.inputs
    label_name = scope.declare_local_variable(
        'label', operator.inputs[0].type.__class__())
    this_operator.outputs.append(label_name)
    apply_identity(scope, label_name.full_name,
                   operator.outputs[0].full_name, container)


register_converter('SklearnRANSACRegressor',
                   convert_sklearn_ransac_regressor)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_cast
from ..common.data_types import (
    Int64TensorType, FloatTensorType, DoubleTensorType,
    StringTensorType, guess_proto_type)


def concatenate_variables(scope, variables, container, main_type=None):
    """
    This function allocate operators to from a float tensor by concatenating
    all input variables. Notice that if all integer inputs would be converted
    to floats before concatenation.
    """
    if main_type is None:
        main_type = variables[0].type.__class__

    # Check if it's possible to concatenate those inputs.
    type_set = set(type(variable.type) for variable in variables)
    number_type_set = {FloatTensorType, Int64TensorType, DoubleTensorType,
                       StringTensorType}
    if any(itype not in number_type_set for itype in type_set):
        raise RuntimeError('Numerical tensor(s) and string tensor(s) '
                           'cannot be concatenated.')
    # input variables' names we want to concatenate
    input_names = []
    # dimensions of the variables that is going to be concatenated
    input_dims = []

    # Collect input variable names and do cast if needed
    for variable in variables:
        if not isinstance(variable.type, main_type):
            proto_type = guess_proto_type(main_type())
            new_name = scope.get_unique_variable_name('cast')
            apply_cast(scope, variable.full_name, new_name,
                       container, to=proto_type)
            input_names.append(new_name)
        else:
            input_names.append(variable.full_name)
        # We assume input variables' shape are [1, C_1], ..., [1, C_n],
        # if there are n inputs.
        input_dims.append(variable.type.shape[1])

    if len(input_names) == 1:
        # No need to concatenate tensors if there is only one input
        return input_names[0]

    # To combine all inputs, we need a FeatureVectorizer
    op_type = 'FeatureVectorizer'
    attrs = {
        'name': scope.get_unique_operator_name(op_type),
        'inputdimensions': input_dims}
    # Create a variable name to capture feature vectorizer's output
    concatenated_name = scope.get_unique_variable_name('concatenated')
    # Set up our FeatureVectorizer
    container.add_node(op_type, input_names, concatenated_name,
                       op_domain='ai.onnx.ml', **attrs)

    return concatenated_name



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator
try:
    from sklearn.utils.validation import _deprecate_positional_args
except ImportError:
    def _deprecate_positional_args(x): return x  # noqa


class WOETransformer(TransformerMixin, BaseEstimator):

    """
    This transformer cannot be trained. It takes a list of intervals,
    one list per columns, and returns for every feature the list
    of intervals this features falls into. This transformer is close
    to `WOEEncoder <https://contrib.scikit-learn.org/category_encoders/
    woe.html?highlight=woe#category_encoders.woe.WOEEncoder>`_.
    *sklearn-onnx* only converts *scikit-learn* models but this class
    can be used as an example for other converters converting similar
    transforms.

    Parameters
    ----------
    intervals : list of list of tuples,
        every tuple is an interval, one list per column,
        a list can be replaced with constant `'passthrough'`
        which leaves the features untouched.
    weights : list of if list of numerical values
        by default, every interval is weights by 1,
        but it is possible to give them a different value,
        it does not apply on *passthrough* columns.
    onehot : boolean
        if true, the transform creates a column for each interval,
        otherwise, every column sticks into one column but labels
        should be distinct for each interval in the same column

    An interval is defined with four values `(a, b, False, True)`:

    * `(a, b, False, False)` means `]a, b[`
    * `(a, b, False, True)` means `]a, b]`
    * `(a, b, True, False)` means `[a, b[`
    * `(a, b, True, True)` means `[a, b]`

    Boolean defines if the extremity belongs to the interval or not.
    By default `(a, b)` is equivalent to `(a, b, False, True)`.
    Method fit adds attributes `intervals_` and `weights_`.
    """

    @_deprecate_positional_args
    def __init__(self, intervals=None, weights=None, onehot=True):
        self.intervals = intervals
        self.weights = weights
        self.onehot = onehot

    def fit(self, X, y=None, sample_weight=None):
        """
        Does nothing except checking *dtype* may be applied.
        """
        self.intervals_ = []
        self.weights_ = []
        dim = 0
        self.indices_ = []
        n_cols = X.shape[1] if X is not None else len(self.intervals)
        for i in range(n_cols):
            if i >= len(self.intervals):
                self.intervals_.append(None)
                self.weights_.append(None)
                self.indices_.append((dim, dim + 1))
                dim += 1
                continue
            intervals = self.intervals[i]
            if intervals == 'passthrough':
                self.intervals_.append(None)
                self.weights_.append(None)
                self.indices_.append((dim, dim + 1))
                dim += 1
                continue
            if not isinstance(intervals, list):
                raise TypeError(
                    "Intervals for column %d must be a list not %r."
                    "" % (i, intervals))
            inlist = []
            inweight = []
            for index, interval in enumerate(intervals):
                if not isinstance(interval, tuple):
                    raise TypeError(
                        "Interval %d is not a tuple but %r." % (i, interval))
                if len(interval) < 2:
                    raise ValueError(
                        "Interval %d should have at least two values "
                        "%r." % interval)
                res = []
                for j in range(0, 2):
                    try:
                        fv = float(interval[j])
                    except (TypeError, ValueError) as e:
                        raise TypeError(
                            "Value at index %d in %r must be a float."
                            "" % (j, interval)) from e
                    res.append(fv)
                if len(interval) >= 3:
                    if not isinstance(interval[2], bool):
                        raise TypeError(
                            "Value at index %i in %r must be a boolean."
                            "" % (2, interval))
                    res.append(interval[2])
                else:
                    res.append(False)
                if len(interval) >= 4:
                    if not isinstance(interval[3], bool):
                        raise TypeError(
                            "Value at index %i in %r must be a boolean."
                            "" % (3, interval))
                    res.append(interval[3])
                else:
                    res.append(True)
                inlist.append(tuple(res))
                if (self.weights is None or i >= len(self.weights) or
                        index >= len(self.weights[i])):
                    inweight.append(1)
                else:
                    inweight.append(self.weights[i][index])

            self.intervals_.append(inlist)
            self.weights_.append(inweight)
            self.indices_.append((dim, dim + len(inlist)))
            dim += len(inlist)

        self.n_dims_ = dim
        return self

    def _transform_column(self, X, column_index):
        col = X[:, column_index]
        intervals = self.intervals_[column_index]
        if intervals is None:
            if self.onehot:
                return col.reshape((-1, 1))
            return col
        res = np.zeros((X.shape[0], len(intervals)), dtype=X.dtype)
        for i, interval in enumerate(intervals):
            if interval[2]:
                left = col >= interval[0]
            else:
                left = col > interval[0]
            if interval[3]:
                right = col <= interval[1]
            else:
                right = col < interval[1]
            res[:, i] = ((left * right).astype(X.dtype) *
                         self.weights_[column_index][i])
        if self.onehot:
            return res
        return res.sum(axis=1, keepdims=0)

    def transform(self, X, y=None):
        """
        Applies the transformation.
        """
        if self.onehot:
            res = np.zeros((X.shape[0], self.n_dims_), dtype=X.dtype)
        else:
            res = np.zeros(X.shape, dtype=X.dtype)
        for i in range(X.shape[1]):
            a, b = self.indices_[i]
            if self.onehot:
                res[:, a: b] = self._transform_column(X, i)
            else:
                res[:, i] = self._transform_column(X, i)
        return res

    def get_feature_names(self):
        """
        Returns the features names.
        """
        names = []
        for i, intervals in enumerate(self.intervals_):
            if intervals is None:
                names.append("X%d" % i)
                continue
            for interval in intervals:
                name = [
                    "[" if interval[2] else "]",
                    str(interval[0]), ",", str(interval[1]),
                    "]" if interval[3] else "["]
                names.append("".join(name))
        return names

    def _decision_thresholds(self, add_index=False):
        "Returns all decision thresholds."
        extremities = []
        for intervals in self.intervals_:
            if intervals is None:
                extremities.append(None)
                continue
            thresholds = []
            for index, interval in enumerate(intervals):
                if add_index:
                    thresholds.append((interval[0], not interval[2], index))
                    thresholds.append((interval[1], interval[3], index))
                else:
                    thresholds.append((interval[0], not interval[2]))
                    thresholds.append((interval[1], interval[3]))
            extremities.append(list(sorted(set(thresholds))))
        return extremities



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_identity
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_identity(scope: Scope, operator: Operator,
                             container: ModelComponentContainer):
    apply_identity(
        scope, operator.inputs[0].full_name,
        operator.outputs[0].full_name, container,
        operator_name=scope.get_unique_operator_name('CIdentity'))


register_converter('SklearnIdentity', convert_sklearn_identity)



# SPDX-License-Identifier: Apache-2.0


from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType
)
from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_k_bins_discretiser(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    M = operator.inputs[0].get_first_dimension()
    model = operator.raw_operator
    N = len(model.n_bins_) if model.encode == 'ordinal' else sum(model.n_bins_)
    operator.outputs[0].type.shape = [M, N]


register_shape_calculator('SklearnKBinsDiscretizer',
                          calculate_sklearn_k_bins_discretiser)



# SPDX-License-Identifier: Apache-2.0


from onnx.helper import make_tensor
from onnx import TensorProto
from ..proto import onnx_proto
from ..common._apply_operation import apply_concat, apply_cast
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_polynomial_features(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    op = operator.raw_operator
    transformed_columns = [None] * (op.n_output_features_)

    if hasattr(op, '_min_degree'):
        # scikit-learn >= 1.0
        combinations = op._combinations(op.n_input_features_,
                                        op._min_degree,
                                        op._max_degree,
                                        op.interaction_only,
                                        op.include_bias)
    else:
        combinations = op._combinations(op.n_input_features_, op.degree,
                                        op.interaction_only,
                                        op.include_bias)

    unit_name = None
    last_feat = None
    for i, comb in enumerate(combinations):
        if len(comb) == 0:
            unit_name = scope.get_unique_variable_name('unit')
            transformed_columns[i] = unit_name
        else:
            comb_name = scope.get_unique_variable_name('comb')
            col_name = scope.get_unique_variable_name('col')
            prod_name = scope.get_unique_variable_name('prod')

            container.add_initializer(comb_name, onnx_proto.TensorProto.INT64,
                                      [len(comb)], list(comb))

            container.add_node(
                'ArrayFeatureExtractor',
                [operator.inputs[0].full_name, comb_name], col_name,
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
                op_domain='ai.onnx.ml')
            reduce_prod_input = col_name
            if (operator.inputs[0].type._get_element_onnx_type()
                    == onnx_proto.TensorProto.INT64):
                float_col_name = scope.get_unique_variable_name('col')

                apply_cast(scope, col_name, float_col_name, container,
                           to=onnx_proto.TensorProto.FLOAT)
                reduce_prod_input = float_col_name

            container.add_node(
                'ReduceProd', reduce_prod_input, prod_name,
                axes=[1], name=scope.get_unique_operator_name('ReduceProd'))
            transformed_columns[i] = prod_name
            last_feat = prod_name

    if unit_name is not None:
        shape_name = scope.get_unique_variable_name('shape')
        container.add_node('Shape', last_feat, shape_name)
        container.add_node('ConstantOfShape', shape_name, unit_name,
                           value=make_tensor(
                               'ONE', TensorProto.FLOAT, [1], [1.]),
                           op_version=9)

    if (operator.inputs[0].type._get_element_onnx_type()
            == onnx_proto.TensorProto.INT64):
        concat_result_name = scope.get_unique_variable_name('concat_result')

        apply_concat(scope, [t for t in transformed_columns],
                     concat_result_name, container, axis=1)
        apply_cast(scope, concat_result_name,
                   operator.outputs[0].full_name, container,
                   to=onnx_proto.TensorProto.INT64)
    else:
        apply_concat(scope, [t for t in transformed_columns],
                     operator.outputs[0].full_name, container, axis=1)


register_converter('SklearnPolynomialFeatures',
                   convert_sklearn_polynomial_features)



# SPDX-License-Identifier: Apache-2.0


from .cast_transformer import CastTransformer  # noqa
from .cast_regressor import CastRegressor  # noqa
from .replace_transformer import ReplaceTransformer  # noqa
from .sklearn_text import TraceableCountVectorizer, TraceableTfidfVectorizer  # noqa
from .woe_transformer import WOETransformer  # noqa



# SPDX-License-Identifier: Apache-2.0

import numpy as np
try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None
from scipy.sparse import coo_matrix
from ..proto import TensorProto, ValueInfoProto
from ..common._topology import Variable
from ..common.data_types import (
    _guess_numpy_type,
    _guess_type_proto,
    BooleanTensorType,
    DataType,
    DoubleType,
    DoubleTensorType,
    FloatType,
    FloatTensorType,
    Int64Type,
    Int64TensorType, Int32TensorType,
    StringTensorType)
from ..common.data_types import (
    Int8TensorType, UInt8TensorType,
    UInt8Type, Int8Type)


def _guess_type(given_type):
    """
    Returns the proper type of an input.
    """
    def _guess_dim(value):
        if value == 0:
            return None
        return value

    if isinstance(given_type, (np.ndarray, coo_matrix)):
        shape = list(given_type.shape)
        if len(shape) == 0:
            # a number
            return _guess_numpy_type(given_type.dtype, tuple())
        shape[0] = None
        return _guess_numpy_type(given_type.dtype, shape)
    if isinstance(given_type, (FloatTensorType, Int64TensorType,
                               Int32TensorType, StringTensorType,
                               BooleanTensorType, DoubleTensorType,
                               Int8TensorType, UInt8TensorType)):
        return given_type
    if isinstance(given_type, Variable):
        return given_type.type
    if isinstance(given_type, DataType):
        return given_type
    if isinstance(given_type, TensorProto):
        return _guess_type_proto(given_type.data_type,
                                 given_type.dims)
    if isinstance(given_type, ValueInfoProto):
        ttype = given_type.type.tensor_type
        dims = [_guess_dim(ttype.shape.dim[i].dim_value)
                for i in range(len(ttype.shape.dim))]
        return _guess_type_proto(ttype.elem_type, dims)
    if isinstance(given_type, np.int64):
        return Int64Type()
    if isinstance(given_type, np.float32):
        return FloatType()
    if isinstance(given_type, np.float64):
        return DoubleType()
    if isinstance(given_type, np.int8):
        return Int8Type()
    if isinstance(given_type, np.uint8):
        return UInt8Type()
    if given_type.__class__.__name__.endswith("Categorical"):
        # pandas Categorical without important pandas
        return Int64TensorType()
    raise NotImplementedError(
        "Unsupported type '{}'. You may raise an issue "
        "at https://github.com/onnx/sklearn-onnx/issues."
        "".format(type(given_type)))


def guess_initial_types(X, initial_types):
    if X is None and initial_types is None:
        raise NotImplementedError("Initial types must be specified.")
    if initial_types is None:
        if isinstance(X, np.ndarray):
            X = X[:1]
            gt = _guess_type(X)
            initial_types = [('X', gt)]
        elif DataFrame is not None and isinstance(X, DataFrame):
            X = X[:1]
            initial_types = []
            for c in X.columns:
                if isinstance(X[c].values[0], (str, np.str_)):
                    g = StringTensorType()
                else:
                    g = _guess_type(X[c].values)
                g.shape = [None, 1]
                initial_types.append((c, g))
        elif isinstance(X, list):
            initial_types = X
        else:
            raise TypeError(
                "Unexpected type %r, unable to guess type." % type(X))
    return initial_types



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from .._supported_operators import sklearn_operator_name_map
from ..common.data_types import Int64TensorType
from ..common._apply_operation import (
    apply_cast, apply_concat,
    apply_div, apply_reshape)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def _calculate_proba(scope, operator, container, model):
    """
    This function calculates class probability scores for
    BaggingClassifier.
    """
    final_proba_name = operator.outputs[1].full_name
    proba_list = []
    options = container.get_options(model, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']
    has_proba = (hasattr(model.estimators_[0], 'predict_proba')
                 or (use_raw_scores and hasattr(
                     model.estimators_[0], 'decision_function')))
    for index, estimator in enumerate(model.estimators_):
        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        if container.has_options(estimator, 'raw_scores'):
            container.add_options(
                id(estimator), {'raw_scores': use_raw_scores})
            scope.add_options(id(estimator), {'raw_scores': use_raw_scores})

        label_name = scope.declare_local_variable(
            'label_%d' % index, Int64TensorType())
        proba_name = scope.declare_local_variable(
            'proba_%d' % index, operator.inputs[0].type.__class__())

        features = model.estimators_features_[index]
        if (len(features) == model.n_features_ and
                list(features) == list(range(model.n_features_))):
            this_operator.inputs = operator.inputs
        else:
            # subset of features
            feat_name = scope.declare_local_variable(
                'fsel_%d' % index, operator.inputs[0].type.__class__())
            index_name = scope.get_unique_variable_name(
                'index_name_%d' % index)
            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64,
                (len(features), ), list(features))
            container.add_node(
                'Gather', [operator.inputs[0].full_name, index_name],
                [feat_name.full_name],
                name=scope.get_unique_operator_name('GatherBG'), axis=1)
            this_operator.inputs.append(feat_name)

        this_operator.outputs.append(label_name)
        this_operator.outputs.append(proba_name)
        proba_output_name = (proba_name.onnx_name if has_proba
                             else label_name.onnx_name)
        reshape_dim_val = len(model.classes_) if has_proba else 1
        reshaped_proba_name = scope.get_unique_variable_name('reshaped_proba')
        apply_reshape(scope, proba_output_name, reshaped_proba_name,
                      container, desired_shape=(1, -1, reshape_dim_val))
        proba_list.append(reshaped_proba_name)
    merged_proba_name = scope.get_unique_variable_name('merged_proba')
    apply_concat(scope, proba_list,
                 merged_proba_name, container, axis=0)
    if has_proba:
        container.add_node('ReduceMean', merged_proba_name,
                           final_proba_name,
                           name=scope.get_unique_operator_name('ReduceMean'),
                           axes=[0], keepdims=0)
    else:
        n_estimators_name = scope.get_unique_variable_name('n_estimators')
        class_labels_name = scope.get_unique_variable_name('class_labels')
        equal_result_name = scope.get_unique_variable_name('equal_result')
        cast_output_name = scope.get_unique_variable_name('cast_output')
        reduced_proba_name = scope.get_unique_variable_name('reduced_proba')

        container.add_initializer(
            n_estimators_name, onnx_proto.TensorProto.FLOAT, [],
            [len(model.estimators_)])
        container.add_initializer(
            class_labels_name, onnx_proto.TensorProto.INT64,
            [1, 1, len(model.estimators_[0].classes_)],
            model.estimators_[0].classes_)

        container.add_node('Equal', [class_labels_name, merged_proba_name],
                           equal_result_name,
                           name=scope.get_unique_operator_name('Equal'))
        apply_cast(scope, equal_result_name, cast_output_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', cast_output_name, reduced_proba_name,
                name=scope.get_unique_operator_name('ReduceSum'),
                axes=[0], keepdims=0)
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [cast_output_name, axis_name],
                reduced_proba_name, keepdims=0,
                name=scope.get_unique_operator_name('ReduceSum'))
        apply_div(scope, [reduced_proba_name, n_estimators_name],
                  final_proba_name, container, broadcast=1)
    return final_proba_name


def convert_sklearn_bagging_classifier(scope: Scope, operator: Operator,
                                       container: ModelComponentContainer):
    """
    Converter for BaggingClassifier.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))['nocl']:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__))

    bagging_op = operator.raw_operator
    classes = bagging_op.classes_
    output_shape = (-1,)
    classes_name = scope.get_unique_variable_name('classes')
    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')
    class_type = onnx_proto.TensorProto.STRING

    if np.issubdtype(bagging_op.classes_.dtype, np.floating):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(bagging_op.classes_.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    proba_name = _calculate_proba(scope, operator, container, bagging_op)
    container.add_node(
        'ArgMax', proba_name, argmax_output_name,
        name=scope.get_unique_operator_name('ArgMax'), axis=1)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    if class_type == onnx_proto.TensorProto.INT32:
        cast_result_name = scope.get_unique_variable_name('cast_result')
        reshaped_result_name = scope.get_unique_variable_name(
            'reshaped_result')
        apply_cast(scope, array_feature_extractor_result_name,
                   cast_result_name, container,
                   to=onnx_proto.TensorProto.INT64)
        apply_reshape(scope, cast_result_name, reshaped_result_name,
                      container, desired_shape=output_shape)
        apply_cast(scope, reshaped_result_name, operator.outputs[0].full_name,
                   container, to=onnx_proto.TensorProto.INT64)
    else:  # string labels
        apply_reshape(scope, array_feature_extractor_result_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=output_shape)


def convert_sklearn_bagging_regressor(scope: Scope, operator: Operator,
                                      container: ModelComponentContainer):
    """
    Converter for BaggingRegressor.
    """
    bagging_op = operator.raw_operator
    proba_list = []
    for index, estimator in enumerate(bagging_op.estimators_):
        op_type = sklearn_operator_name_map[type(estimator)]
        this_operator = scope.declare_local_operator(op_type, estimator)

        features = bagging_op.estimators_features_[index]
        if (len(features) == bagging_op.n_features_ and
                list(features) == list(range(bagging_op.n_features_))):
            this_operator.inputs = operator.inputs
        else:
            # subset of features
            feat_name = scope.declare_local_variable(
                'fsel_%d' % index, operator.inputs[0].type.__class__())
            index_name = scope.get_unique_variable_name('index_name')
            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64,
                (len(features), ), list(features))
            container.add_node(
                'Gather', [operator.inputs[0].full_name, index_name],
                [feat_name.full_name],
                name=scope.get_unique_operator_name('GatherBG'), axis=1)
            this_operator.inputs.append(feat_name)

        label_name = scope.declare_local_variable(
            'variable_%d' % index, this_operator.inputs[0].type.__class__())
        this_operator.outputs.append(label_name)
        reshaped_proba_name = scope.get_unique_variable_name('reshaped_proba')
        apply_reshape(scope, label_name.onnx_name, reshaped_proba_name,
                      container, desired_shape=(1, -1, 1))
        proba_list.append(reshaped_proba_name)

    merged_proba_name = scope.get_unique_variable_name('merged_proba')
    apply_concat(scope, proba_list,
                 merged_proba_name, container, axis=0)
    container.add_node('ReduceMean', merged_proba_name,
                       operator.outputs[0].full_name,
                       name=scope.get_unique_operator_name('ReduceMean'),
                       axes=[0], keepdims=0)


register_converter('SklearnBaggingClassifier',
                   convert_sklearn_bagging_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})
register_converter('SklearnBaggingRegressor',
                   convert_sklearn_bagging_regressor)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType
)
from ..common.utils import check_input_and_output_types


def calculate_sklearn_kmeans_output_shapes(operator):
    check_input_and_output_types(
        operator,
        good_input_types=[Int64TensorType, FloatTensorType, DoubleTensorType],
        good_output_types=[Int64TensorType, FloatTensorType, DoubleTensorType])
    if len(operator.inputs) != 1:
        raise RuntimeError("Only one input vector is allowed for KMeans.")
    if len(operator.outputs) != 2:
        raise RuntimeError("Two outputs are expected for KMeans.")

    variable = operator.inputs[0]
    N = variable.get_first_dimension()
    op = operator.raw_operator
    operator.outputs[0].type.shape = [N]
    operator.outputs[1].type.shape = [N, op.n_clusters]


register_shape_calculator('SklearnKMeans',
                          calculate_sklearn_kmeans_output_shapes)
register_shape_calculator('SklearnMiniBatchKMeans',
                          calculate_sklearn_kmeans_output_shapes)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_add, apply_cast, apply_div, apply_exp,
    apply_log, apply_mul, apply_pow, apply_sub, apply_reshape,
    apply_transpose,
)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_numpy_type,
    guess_proto_type)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.utils_classifier import get_label_classes


def _joint_log_likelihood_bernoulli(
        scope, container, input_name, feature_log_prob_name,
        class_log_prior_name, binarize, feature_count, proto_dtype,
        sum_result_name):
    """
    Calculate joint log likelihood for Bernoulli Naive Bayes model.
    """
    constant_name = scope.get_unique_variable_name('constant')
    exp_result_name = scope.get_unique_variable_name('exp_result')
    sub_result_name = scope.get_unique_variable_name('sub_result')
    neg_prob_name = scope.get_unique_variable_name('neg_prob')
    sum_neg_prob_name = scope.get_unique_variable_name('sum_neg_prob')
    difference_matrix_name = scope.get_unique_variable_name(
        'difference_matrix')
    dot_prod_name = scope.get_unique_variable_name('dot_prod')
    partial_sum_result_name = scope.get_unique_variable_name(
        'partial_sum_result')
    # Define constant slightly greater than 1 to avoid log 0
    # scenarios when calculating log (1 - x) and x=1 in line 70
    container.add_initializer(constant_name, proto_dtype, [], [1.000000001])

    if binarize is not None:
        threshold_name = scope.get_unique_variable_name('threshold')
        condition_name = scope.get_unique_variable_name('condition')
        cast_values_name = scope.get_unique_variable_name('cast_values')
        zero_tensor_name = scope.get_unique_variable_name('zero_tensor')
        binarised_input_name = scope.get_unique_variable_name(
            'binarised_input')
        num_features = feature_count.shape[1]

        container.add_initializer(threshold_name, proto_dtype,
                                  [1], [binarize])
        container.add_initializer(
            zero_tensor_name,
            proto_dtype, [1, num_features],
            np.zeros((1, num_features)).ravel())

        container.add_node(
            'Greater', [input_name, threshold_name],
            condition_name, name=scope.get_unique_operator_name('Greater'),
            op_version=9)
        apply_cast(scope, condition_name, cast_values_name, container,
                   to=proto_dtype)
        apply_add(scope, [zero_tensor_name, cast_values_name],
                  binarised_input_name, container, broadcast=1)
        input_name = binarised_input_name

    apply_exp(scope, feature_log_prob_name, exp_result_name, container)
    apply_sub(scope, [constant_name, exp_result_name], sub_result_name,
              container, broadcast=1)
    apply_log(scope, sub_result_name, neg_prob_name, container)
    if container.target_opset < 13:
        container.add_node('ReduceSum', neg_prob_name,
                           sum_neg_prob_name, axes=[0],
                           name=scope.get_unique_operator_name('ReduceSum'))
    else:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [0])
        container.add_node(
            'ReduceSum', [neg_prob_name, axis_name], sum_neg_prob_name,
            name=scope.get_unique_operator_name('ReduceSum'))
    apply_sub(scope, [feature_log_prob_name, neg_prob_name],
              difference_matrix_name, container)
    container.add_node(
        'MatMul', [input_name, difference_matrix_name],
        dot_prod_name, name=scope.get_unique_operator_name('MatMul'))

    apply_add(scope, [dot_prod_name, sum_neg_prob_name],
              partial_sum_result_name, container)
    apply_add(scope, [partial_sum_result_name, class_log_prior_name],
              sum_result_name, container)
    return sum_result_name


def _joint_log_likelihood_gaussian(
        scope, container, input_name, model, proto_dtype, sum_result_name):
    """
    Calculate joint log likelihood for Gaussian Naive Bayes model.
    """
    features = model.theta_.shape[1]
    jointi = np.log(model.class_prior_)
    sigma_sum_log = - 0.5 * np.sum(np.log(2. * np.pi * model.sigma_), axis=1)
    theta_name = scope.get_unique_variable_name('theta')
    sigma_name = scope.get_unique_variable_name('sigma')
    sigma_sum_log_name = scope.get_unique_variable_name('sigma_sum_log')
    jointi_name = scope.get_unique_variable_name('jointi')
    exponent_name = scope.get_unique_variable_name('exponent')
    prod_operand_name = scope.get_unique_variable_name('prod_operand')
    reshaped_input_name = scope.get_unique_variable_name('reshaped_input')
    subtracted_input_name = scope.get_unique_variable_name('subtracted_input')
    pow_result_name = scope.get_unique_variable_name('pow_result')
    div_result_name = scope.get_unique_variable_name('div_result')
    reduced_sum_name = scope.get_unique_variable_name('reduced_sum')
    mul_result_name = scope.get_unique_variable_name('mul_result')
    part_log_likelihood_name = scope.get_unique_variable_name(
        'part_log_likelihood')

    theta = model.theta_.reshape((1, -1, features))
    sigma = model.sigma_.reshape((1, -1, features))

    container.add_initializer(theta_name, proto_dtype, theta.shape,
                              theta.ravel())
    container.add_initializer(sigma_name, proto_dtype, sigma.shape,
                              sigma.ravel())
    container.add_initializer(jointi_name, proto_dtype, [1, jointi.shape[0]],
                              jointi)
    container.add_initializer(
        sigma_sum_log_name, proto_dtype,
        [1, sigma_sum_log.shape[0]], sigma_sum_log.ravel())
    container.add_initializer(exponent_name, proto_dtype, [], [2])
    container.add_initializer(prod_operand_name, proto_dtype, [], [0.5])

    apply_reshape(scope, input_name, reshaped_input_name, container,
                  desired_shape=[-1, 1, features])
    apply_sub(scope, [reshaped_input_name, theta_name], subtracted_input_name,
              container, broadcast=1)
    apply_pow(scope, [subtracted_input_name, exponent_name], pow_result_name,
              container, broadcast=1)
    apply_div(scope, [pow_result_name, sigma_name], div_result_name,
              container, broadcast=1)
    if container.target_opset < 13:
        container.add_node('ReduceSum', div_result_name,
                           reduced_sum_name, axes=[2], keepdims=0,
                           name=scope.get_unique_operator_name('ReduceSum'))
    else:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [2])
        container.add_node(
            'ReduceSum', [div_result_name, axis_name], reduced_sum_name,
            keepdims=0, name=scope.get_unique_operator_name('ReduceSum'))
    apply_mul(scope, [reduced_sum_name, prod_operand_name], mul_result_name,
              container, broadcast=1)
    apply_sub(scope, [sigma_sum_log_name, mul_result_name],
              part_log_likelihood_name,
              container, broadcast=1)
    apply_add(scope, [jointi_name, part_log_likelihood_name],
              sum_result_name, container, broadcast=1)
    return sum_result_name


def _joint_log_likelihood_categorical(
        scope, container, input_name, model, sum_result_name):
    """
    Calculate joint log likelihood for Categorical Naive Bayes model.
    """
    jll_list = []
    class_log_prior_name = scope.get_unique_variable_name('class_log_prior')
    summation_jll_name = scope.get_unique_variable_name('summation_jll')

    container.add_initializer(
        class_log_prior_name, onnx_proto.TensorProto.FLOAT,
        model.class_log_prior_.shape, model.class_log_prior_)

    for i in range(model.n_features_):
        feature_index_name = scope.get_unique_variable_name('feature_index')
        indices_name = scope.get_unique_variable_name('indices')
        cast_indices_name = scope.get_unique_variable_name('cast_indices')
        feature_log_proba_name = scope.get_unique_variable_name(
            'feature_log_proba')
        jll_name = scope.get_unique_variable_name('jll')
        transposed_jll_name = scope.get_unique_variable_name('transposed_jll')

        container.add_initializer(
            feature_index_name, onnx_proto.TensorProto.INT64, [], [i])
        container.add_initializer(
            feature_log_proba_name, onnx_proto.TensorProto.FLOAT,
            model.feature_log_prob_[i].shape,
            model.feature_log_prob_[i].ravel())

        container.add_node(
            'ArrayFeatureExtractor', [input_name, feature_index_name],
            indices_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
        apply_cast(scope, indices_name, cast_indices_name,
                   container, to=onnx_proto.TensorProto.INT64)
        container.add_node(
            'ArrayFeatureExtractor',
            [feature_log_proba_name, cast_indices_name],
            jll_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
        apply_transpose(scope, jll_name, transposed_jll_name,
                        container, perm=(1, 0))
        jll_list.append(transposed_jll_name)
    container.add_node('Sum', jll_list,
                       summation_jll_name,
                       name=scope.get_unique_operator_name('Sum'))
    apply_add(scope, [summation_jll_name, class_log_prior_name],
              sum_result_name, container, broadcast=1)
    return sum_result_name


def convert_sklearn_naive_bayes(scope: Scope, operator: Operator,
                                container: ModelComponentContainer):
    # Computational graph:
    #
    # Note: In the following graph, variable names are in lower case
    # characters only and operator names are in upper case characters.
    # We borrow operator names from the official ONNX spec:
    # https://github.com/onnx/onnx/blob/master/docs/Operators.md
    # All variables are followed by their shape in [].
    #
    # Symbols:
    # M: Number of instances
    # N: Number of features
    # C: Number of classes
    # input(or x): input
    # output(or y): output (There are two paths for producing output, one for
    #               string labels and the other one for int labels)
    # output_probability: class probabilties
    # feature_log_prob: Empirical log probability of features given a
    #                   class, P(x_i|y)
    # class_log_prior: Smoothed empirical log probability for each class
    #
    # Multinomial NB
    # Equation:
    #   y = argmax (class_log_prior + X . feature_log_prob^T)
    #
    # Graph:
    #
    #   input [M, N] -> MATMUL <- feature_log_prob.T [N, C]
    #                    |
    #                    V
    #        matmul_result [M, C] -> CAST <- proto_dtype
    #                                |
    #                                V
    #                    cast_result [M, C] -> ADD <- class_log_prior [1, C]
    #                                          |
    #                        .-----------------'
    #                        |
    #                        V
    # sum_result [M, C] -> ARGMAX -> argmax_output [M, 1]
    #                        |
    #                        V
    # classes [C] -----> ARRAYFEATUREEXTRACTOR
    #                        |
    #                        V               (string labels)
    # array_feature_extractor_result [M, 1] --------------------------.
    #           (int labels) |                                        |
    #                        V                                        |
    #              CAST(to=proto_dtype)              |
    #                        |                                        |
    #                        V                                        |
    #                  cast2_result [M, 1]                            |
    #                        |                                        |
    #                        V                                        |
    # output_shape [1] -> RESHAPE                                     |
    #                        |                                        |
    #                        V                                        V
    #                       reshaped_result [M,]            .-----RESHAPE
    #                                   |                   |
    #                                   V                   V
    #  (to=onnx_proto.TensorProto.INT64)CAST --------> output [M,]
    #
    # Bernoulli NB
    # Equation:
    #   y = argmax (class_log_prior + \sum neg_prob
    #               + X . (feature_log_prob - neg_prob))
    #   neg_prob = log( 1 - e ^ feature_log_prob)
    #
    #   Graph:
    #
    #           .---------------------------------------------------------.
    #           |                                                         |
    #  feature_log_prob.T [N, C] -> EXP -> exp_result [N, C]              |
    #                                      |                              |
    #               .----------------------'                              |
    #               |                                                     |
    #               V                                                     V
    #  constant -> SUB -> sub_result [N, C] -> LOG -> neg_prob [N, C] -> SUB
    #                                                   |                 |
    #                                         .---------'       .---------'
    #                                         |                 |
    #                                         V                 V
    #  .----------- sum_neg_prob [1, C] <- REDUCE_SUM  difference_matrix [N, C]
    #  |                                                        |
    #  |                     .----------------------------------'
    #  |                     |
    #  |                     V
    #  |    input [M, N] -> MATMUL -> dot_product [M, C]
    #  |                                       |
    #  |                                       V
    #  '------------------------------------> ADD
    #                                          |
    #                                          V
    #  class_log_prior [1, C] -> ADD <- partial_sum_result [M, C]
    #                            |
    #                            V
    #                   sum_result [M, C] -> ARGMAX -> argmax_output [M, 1]
    #                                                            |
    #                                                            V
    #                              classes [C] -------> ARRAYFEATUREEXTRACTOR
    #                                                            |
    #                       .------------------------------------'
    #                       |
    #                       V                (string labels)
    #  array_feature_extractor_result [M, 1] ----------------.
    #          (int labels) |                                |
    #                       V                                |
    #   CAST(to=proto_dtype)                |
    #                       |                                |
    #                       V                                |
    #                cast2_result [M, 1]                     |
    #                       |                                |
    #                       V                                |
    #  output_shape [1] -> RESHAPE                           |
    #                          |                             |
    #                          V                             V
    #                       reshaped_result [M,]           RESHAPE
    #                                   |                    |
    #                                   V                    |
    # (to=onnx_proto.TensorProto.INT64)CAST -> output [M,] <-'
    #
    #
    # If model's binarize attribute is not null, then input of
    # Bernoulli NB is produced by the following graph:
    #
    #    input [M, N] -> GREATER <- threshold [1]
    #       |              |
    #       |              V
    #       |       condition [M, N] -> CAST(to=proto_dtype)
    #       |                             |
    #       |                             V
    #       |                          cast_values [M, N]
    #       |                                   |
    #       V                                   V
    #   CONSTANT_LIKE -> zero_tensor [M, N] -> ADD
    #                                           |
    #                                           V
    #               input [M, N] <- binarised_input [M, N]
    #
    # Sub-graph for probability calculation common to both Multinomial
    # and Bernoulli Naive Bayes
    #
    #  sum_result [M, C] -> REDUCELOGSUMEXP -> reduce_log_sum_exp_result [M,]
    #         |                                                   |
    #         |                                                   V
    #         |                           log_prob_shape [2] -> RESHAPE
    #         |                                                   |
    #         '------------> SUB <-- reshaped_log_prob [M, 1] <---'
    #                         |
    #                         V
    #                     log_prob [M, C] -> EXP -> prob_tensor [M, C] -.
    #                                                                   |
    #         output_probability [M, C] <- ZIPMAP <---------------------'
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    float_dtype = dtype
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    nb_op = operator.raw_operator
    classes = get_label_classes(scope, nb_op)
    output_shape = (-1,)

    sum_result_name = scope.get_unique_variable_name('sum_result')
    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    cast2_result_name = scope.get_unique_variable_name('cast2_result')
    reshaped_result_name = scope.get_unique_variable_name('reshaped_result')
    classes_name = scope.get_unique_variable_name('classes')
    reduce_log_sum_exp_result_name = scope.get_unique_variable_name(
        'reduce_log_sum_exp_result')
    log_prob_name = scope.get_unique_variable_name('log_prob')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')

    class_type = onnx_proto.TensorProto.STRING
    if np.issubdtype(classes.dtype, np.floating):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    if operator.type not in ('SklearnCategoricalNB', 'SklearnGaussianNB'):
        class_log_prior_name = scope.get_unique_variable_name(
            'class_log_prior')
        feature_log_prob_name = scope.get_unique_variable_name(
            'feature_log_prob')

        class_log_prior = nb_op.class_log_prior_.astype(
            float_dtype).reshape((1, -1))
        feature_log_prob = nb_op.feature_log_prob_.T.astype(float_dtype)

        container.add_initializer(
            feature_log_prob_name, proto_dtype,
            feature_log_prob.shape, feature_log_prob.flatten())
        container.add_initializer(
            class_log_prior_name, proto_dtype,
            class_log_prior.shape, class_log_prior.flatten())

    input_name = operator.inputs[0].full_name
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=proto_dtype)
        input_name = cast_input_name

    if operator.type == 'SklearnBernoulliNB':
        sum_result_name = _joint_log_likelihood_bernoulli(
            scope, container, input_name, feature_log_prob_name,
            class_log_prior_name, nb_op.binarize, nb_op.feature_count_,
            proto_dtype, sum_result_name)
    elif operator.type == 'SklearnGaussianNB':
        sum_result_name = _joint_log_likelihood_gaussian(
            scope, container, input_name, nb_op,
            proto_dtype, sum_result_name)
    elif operator.type == 'SklearnCategoricalNB':
        sum_result_name = _joint_log_likelihood_categorical(
            scope, container, input_name, nb_op, sum_result_name)
    else:
        # MultinomialNB or ComplementNB
        matmul_result_name = (
            scope.get_unique_variable_name('matmul_result')
            if operator.type == 'SklearnMultinomialNB' or len(classes) == 1
            else sum_result_name)

        container.add_node(
            'MatMul', [input_name, feature_log_prob_name],
            matmul_result_name, name=scope.get_unique_operator_name('MatMul'))
        if operator.type == 'SklearnMultinomialNB' or len(classes) == 1:
            apply_add(scope, [matmul_result_name, class_log_prior_name],
                      sum_result_name, container, broadcast=1)

    container.add_node('ArgMax', sum_result_name,
                       argmax_output_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)

    # Calculation of class probability
    log_prob_shape = [-1, 1]

    reshaped_log_prob_name = scope.get_unique_variable_name(
        'reshaped_log_prob')

    container.add_node('ReduceLogSumExp', sum_result_name,
                       reduce_log_sum_exp_result_name,
                       name=scope.get_unique_operator_name('ReduceLogSumExp'),
                       axes=[1], keepdims=0)
    apply_reshape(scope, reduce_log_sum_exp_result_name,
                  reshaped_log_prob_name, container,
                  desired_shape=log_prob_shape)
    apply_sub(scope, [sum_result_name, reshaped_log_prob_name], log_prob_name,
              container, broadcast=1)
    apply_exp(scope, log_prob_name, operator.outputs[1].full_name, container)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    # Reshape op does not seem to handle INT64 tensor even though it is
    # listed as one of the supported types in the doc, so Cast was
    # required here.
    if class_type == onnx_proto.TensorProto.INT32:
        apply_cast(scope, array_feature_extractor_result_name,
                   cast2_result_name, container,
                   to=proto_dtype)
        apply_reshape(scope, cast2_result_name, reshaped_result_name,
                      container, desired_shape=output_shape)
        apply_cast(scope, reshaped_result_name, operator.outputs[0].full_name,
                   container, to=onnx_proto.TensorProto.INT64)
    else:  # string labels
        apply_reshape(scope, array_feature_extractor_result_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=output_shape)


register_converter('SklearnBernoulliNB', convert_sklearn_naive_bayes,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnCategoricalNB', convert_sklearn_naive_bayes,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnComplementNB', convert_sklearn_naive_bayes,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnGaussianNB', convert_sklearn_naive_bayes,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnMultinomialNB', convert_sklearn_naive_bayes,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})



# SPDX-License-Identifier: Apache-2.0


# To register shape calculators for scikit-learn operators,
# import associated modules here.
from . import array_feature_extractor
from . import cast_op
from . import class_labels
from . import concat
from . import cross_decomposition
from . import dict_vectorizer
from . import ensemble_shapes
from . import flatten
from . import function_transformer
from . import gaussian_process
from . import grid_search_cv
from . import identity
from . import imputer
from . import isolation_forest
from . import kernel_pca
from . import k_bins_discretiser
from . import k_means
from . import label_binariser
from . import label_encoder
from . import linear_classifier
from . import linear_regressor
from . import local_outlier_factor
from . import mixture
from . import multioutput
from . import nearest_neighbours
from . import one_hot_encoder
from . import ordinal_encoder
from . import one_vs_one_classifier
from . import one_vs_rest_classifier
from . import ovr_decision_function
from . import pipelines
from . import polynomial_features
from . import power_transformer
from . import quadratic_discriminant_analysis
from . import random_projection
from . import random_trees_embedding
from . import replace_op
from . import scaler
from . import sequence
from . import sgd_oneclass_svm
from . import svd
from . import support_vector_machines
from . import text_vectorizer
from . import tfidf_transformer
from . import voting_classifier
from . import voting_regressor
from . import zip_map

__all__ = [
    array_feature_extractor,
    cast_op,
    class_labels,
    concat,
    cross_decomposition,
    dict_vectorizer,
    ensemble_shapes,
    flatten,
    function_transformer,
    gaussian_process,
    grid_search_cv,
    identity,
    imputer,
    isolation_forest,
    kernel_pca,
    k_bins_discretiser,
    k_means,
    label_binariser,
    label_encoder,
    linear_classifier,
    linear_regressor,
    local_outlier_factor,
    mixture,
    multioutput,
    nearest_neighbours,
    one_hot_encoder,
    ordinal_encoder,
    one_vs_one_classifier,
    one_vs_rest_classifier,
    ovr_decision_function,
    pipelines,
    polynomial_features,
    power_transformer,
    quadratic_discriminant_analysis,
    random_projection,
    random_trees_embedding,
    replace_op,
    scaler,
    sequence,
    sgd_oneclass_svm,
    svd,
    support_vector_machines,
    text_vectorizer,
    tfidf_transformer,
    voting_classifier,
    voting_regressor,
    zip_map,
]



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common.data_types import DoubleTensorType
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .common import concatenate_variables


def convert_sklearn_binarizer(scope: Scope, operator: Operator,
                              container: ModelComponentContainer):
    feature_name = concatenate_variables(scope, operator.inputs, container)

    if isinstance(operator.inputs[0].type, DoubleTensorType):
        name0 = scope.get_unique_variable_name('cst0')
        name1 = scope.get_unique_variable_name('cst1')
        thres = scope.get_unique_variable_name('th')
        container.add_initializer(
            name0, onnx_proto.TensorProto.DOUBLE, [], [0.])
        container.add_initializer(
            name1, onnx_proto.TensorProto.DOUBLE, [], [1.])
        container.add_initializer(
            thres, onnx_proto.TensorProto.DOUBLE, [],
            [float(operator.raw_operator.threshold)])
        binbool = scope.get_unique_variable_name('binbool')
        container.add_node(
            'Less', [feature_name, thres], binbool,
            name=scope.get_unique_operator_name('Less'))
        container.add_node(
            'Where', [binbool, name0, name1], operator.output_full_names,
            name='Where')
        return

    op_type = 'Binarizer'
    attrs = {
        'name': scope.get_unique_operator_name(op_type),
        'threshold': float(operator.raw_operator.threshold)
    }
    container.add_node(op_type, feature_name, operator.output_full_names,
                       op_domain='ai.onnx.ml', **attrs)


register_converter('SklearnBinarizer', convert_sklearn_binarizer)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import (
    check_input_and_output_numbers, check_input_and_output_types)
from ..common.shape_calculator import (
    calculate_linear_regressor_output_shapes,
    calculate_linear_classifier_output_shapes,
    _calculate_linear_classifier_output_shapes)
from ..common.data_types import (
    BooleanTensorType,
    DoubleTensorType,
    FloatTensorType,
    Int64TensorType,
)


def calculate_tree_regressor_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, 1]

    This operator produces a scalar prediction for every example in a
    batch. If the input batch size is N, the output shape may be
    [N, 1].
    """
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=[1, 3])
    check_input_and_output_types(operator, good_input_types=[
        BooleanTensorType, DoubleTensorType,
        FloatTensorType, Int64TensorType])

    N = operator.inputs[0].get_first_dimension()
    if operator.outputs[0].type is None:
        raise RuntimeError(
            "Output type is unknown for operator %r." % operator)
    operator.outputs[0].type.shape = [N, 1]

    # decision_path, decision_leaf
    for n in range(2, len(operator.outputs)):
        if hasattr(operator.raw_operator, 'estimators_'):
            # random forest
            operator.outputs[n].type.shape = [
                N, len(operator.raw_operator.estimators_)]
        else:
            # single tree
            operator.outputs[n].type.shape = [N, 1]


def calculate_tree_classifier_output_shapes(operator):
    _calculate_linear_classifier_output_shapes(operator, True, True)
    N = operator.inputs[0].get_first_dimension()

    # decision_path, decision_leaf
    for n in range(2, len(operator.outputs)):
        if operator.outputs[n].type is None:
            raise RuntimeError(
                "Output type is unknown for operator %r." % operator)
        if hasattr(operator.raw_operator, 'estimators_'):
            # random forest
            operator.outputs[n].type.shape = [
                N, len(operator.raw_operator.estimators_)]
        else:
            # single tree
            operator.outputs[n].type.shape = [N, 1]


register_shape_calculator('SklearnDecisionTreeRegressor',
                          calculate_tree_regressor_output_shapes)
register_shape_calculator('SklearnExtraTreeRegressor',
                          calculate_tree_regressor_output_shapes)
register_shape_calculator('SklearnExtraTreesRegressor',
                          calculate_tree_regressor_output_shapes)
register_shape_calculator('SklearnGradientBoostingRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnHistGradientBoostingRegressor',
                          calculate_linear_regressor_output_shapes)
register_shape_calculator('SklearnRandomForestRegressor',
                          calculate_tree_regressor_output_shapes)

register_shape_calculator('SklearnDecisionTreeClassifier',
                          calculate_tree_classifier_output_shapes)
register_shape_calculator('SklearnExtraTreeClassifier',
                          calculate_tree_classifier_output_shapes)
register_shape_calculator('SklearnExtraTreesClassifier',
                          calculate_tree_classifier_output_shapes)
register_shape_calculator('SklearnGradientBoostingClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnHistGradientBoostingClassifier',
                          calculate_linear_classifier_output_shapes)
register_shape_calculator('SklearnRandomForestClassifier',
                          calculate_tree_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_flatten(scope: Scope, operator: Operator,
                            container: ModelComponentContainer):
    name = scope.get_unique_operator_name('Flatten')
    container.add_node('Flatten', operator.inputs[0].full_name,
                       operator.outputs[0].full_name, name=name,
                       axis=1)


register_converter('SklearnFlatten', convert_sklearn_flatten)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import FloatType, Int64Type, StringType, TensorType
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_flatten(operator):
    check_input_and_output_numbers(operator, output_count_range=1,
                                   input_count_range=1)
    i = operator.inputs[0]
    N = i.get_first_dimension()
    if isinstance(i.type, TensorType):
        if i.type.shape[1] is None:
            C = None
        else:
            C = i.type.shape[1]
    elif isinstance(i.type, (Int64Type, FloatType, StringType)):
        C = 1
    else:
        C = None
    if C is None:
        operator.outputs[0].type.shape = [N, C]
    else:
        operator.outputs[0].type.shape = [N * C]


register_shape_calculator('SklearnFlatten', calculate_sklearn_flatten)



# SPDX-License-Identifier: Apache-2.0

from .utils_checking import check_signature

# This dictionary defines the converters which can be invoked in the
# conversion framework defined in _topology.py. A key in this dictionary
# is an operator's unique ID (e.g., string and type) while the
# associated value is the callable object used to convert the
# operator specified by the key.
_converter_pool = {}


class RegisteredConverter:

    def __init__(self, fct, options):
        self._fct = fct
        self._options = options

    def __call__(self, *args):
        if (len(args) == 3 and
                hasattr(args[2], '_get_allowed_options') and
                hasattr(args[1], 'raw_operator')):
            # Checks that the user did not specify a wrong option.
            if args[1].raw_operator is not None:
                args[2]._get_allowed_options(args[1].raw_operator)
        return self._fct(*args)

    def get_allowed_options(self):
        return self._options


# This dictionary defines the shape calculators which can be invoked in
# the conversion framework defined in _topology.py. A key in this
# dictionary is an operator's unique ID (e.g., string and type) while
# the associated value is the callable object used to infer the output
# shape(s) for the operator specified by the key.
_shape_calculator_pool = {}


def register_converter(operator_name, conversion_function, overwrite=False,
                       options=None):
    """
    :param operator_name: A unique operator ID. It is usually a string
                          but you can use a type as well
    :param conversion_function: A callable object
    :param overwrite: By default, we raise an exception if the caller of
                      this function is trying to assign an existing
                      key (i.e., operator_name) a new value
                      (i.e., conversion_function). Set this flag to True
                      to enable overwriting.
    :param options: supported options for this converter
        (dictionary {name: supported values or None})
    """
    if conversion_function is None:
        raise ValueError(
            "A converter cannot be None for %r." % operator_name)
    if not overwrite and operator_name in _converter_pool:
        raise ValueError('We do not overwrite registered converter '
                         'by default')
    if len(_converter_pool) > 0:
        key = next(iter(_converter_pool))
        check_signature(conversion_function, _converter_pool[key]._fct,
                        skip=('operator', ))
    _converter_pool[operator_name] = RegisteredConverter(
        conversion_function, options)


def get_converter(operator_name):
    if operator_name not in _converter_pool:
        msg = 'Unsupported conversion for operator %s (%d registered)' % (
            operator_name, len(_converter_pool))
        raise ValueError(msg)
    return _converter_pool[operator_name]


def register_shape_calculator(operator_name, calculator_function,
                              overwrite=False):
    """
    :param operator_name: A unique operator ID. It is usually a string
                          but you can use a type as well
    :param calculator_function: A callable object
    :param overwrite: By default, we raise an exception if the caller
                      of this function is trying to assign an existing
                      key (i.e., operator_name) a new value
                      (i.e., calculator_function). Set this flag to True
                      to enable overwriting.
    """
    if calculator_function is None:
        raise ValueError(
            "A shape calculator cannot be None for %r." % operator_name)
    if not overwrite and operator_name in _shape_calculator_pool:
        raise ValueError('We do not overwrite registrated shape calculator '
                         'by default')
    if calculator_function is not None and len(_shape_calculator_pool) > 0:
        key = next(iter(_shape_calculator_pool))
        check_signature(calculator_function, _shape_calculator_pool[key],
                        skip=('operator', ))
    _shape_calculator_pool[operator_name] = calculator_function


def get_shape_calculator(operator_name):
    if operator_name not in _shape_calculator_pool:
        msg = ("Unsupported shape calculator for operator "
               "'%s'." % operator_name)
        raise ValueError(msg)
    return _shape_calculator_pool[operator_name]



# SPDX-License-Identifier: Apache-2.0

import math
import numpy as np
from sklearn.gaussian_process.kernels import (
    Sum, Product, ConstantKernel,
    RBF, DotProduct, ExpSineSquared,
    RationalQuadratic, PairwiseKernel
)
from ..algebra.complex_functions import onnx_squareform_pdist, onnx_cdist
from ..algebra.onnx_ops import (
    OnnxMul, OnnxMatMul, OnnxAdd,
    OnnxTranspose, OnnxDiv, OnnxExp,
    OnnxShape, OnnxSin, OnnxPow,
    OnnxReduceSumApi11, OnnxSqueezeApi11,
    OnnxIdentity, OnnxReduceSumSquare,
    OnnxReduceL2_typed
)
from ..algebra.custom_ops import OnnxCDist
from ..proto.onnx_helper_modified import from_array
try:
    from ..algebra.onnx_ops import OnnxConstantOfShape
except ImportError:
    OnnxConstantOfShape = None


def convert_kernel_diag(kernel, X, output_names=None, dtype=None,
                        optim=None, op_version=None):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if isinstance(kernel, Sum):
        return OnnxAdd(
            convert_kernel_diag(
                kernel.k1, X, dtype=dtype, optim=optim, op_version=op_version),
            convert_kernel_diag(
                kernel.k2, X, dtype=dtype, optim=optim, op_version=op_version),
            output_names=output_names,
            op_version=op_version)

    if isinstance(kernel, Product):
        return OnnxMul(
            convert_kernel_diag(
                kernel.k1, X, dtype=dtype, optim=optim, op_version=op_version),
            convert_kernel_diag(
                kernel.k2, X, dtype=dtype, optim=optim, op_version=op_version),
            output_names=output_names, op_version=op_version)

    if isinstance(kernel, ConstantKernel):
        onnx_zeros = _zero_vector_of_size(X, keepdims=0, dtype=dtype,
                                          op_version=op_version)
        return OnnxAdd(onnx_zeros,
                       np.array([kernel.constant_value],
                                dtype=dtype),
                       output_names=output_names, op_version=op_version)

    if isinstance(kernel, (RBF, ExpSineSquared, RationalQuadratic)):
        onnx_zeros = _zero_vector_of_size(X, keepdims=0, dtype=dtype,
                                          op_version=op_version)
        if isinstance(kernel, RBF):
            return OnnxAdd(onnx_zeros,
                           np.array([1], dtype=dtype),
                           output_names=output_names,
                           op_version=op_version)
        else:
            return OnnxAdd(onnx_zeros, np.array([1], dtype=dtype),
                           output_names=output_names, op_version=op_version)

    if isinstance(kernel, DotProduct):
        t_sigma_0 = py_make_float_array(kernel.sigma_0 ** 2, dtype=dtype)
        return OnnxSqueezeApi11(
            OnnxAdd(OnnxReduceSumSquare(X, axes=[1], op_version=op_version),
                    t_sigma_0, op_version=op_version),
            output_names=output_names, axes=[1],
            op_version=op_version)

    raise RuntimeError("Unable to convert diag method for "
                       "class {}.".format(type(kernel)))


def py_make_float_array(cst, dtype, as_tensor=False):
    if dtype not in (np.float32, np.float64):
        raise TypeError("A float array must be of dtype "
                        "np.float32 or np.float64 ({}).".format(dtype))
    if isinstance(cst, np.ndarray):
        res = cst.astype(dtype)
    elif not isinstance(cst, (int, float, np.float32, np.float64,
                              np.int32, np.int64)):
        raise TypeError("cst must be a number not {}".format(type(cst)))
    else:
        res = np.array([cst], dtype=dtype)
    return from_array(res) if as_tensor else res


def _convert_exp_sine_squared(X, Y, length_scale=1.2, periodicity=1.1,
                              pi=math.pi, dtype=None, optim=None,
                              op_version=None, **kwargs):
    """
    Implements the kernel ExpSineSquared.
    """
    if optim is None:
        dists = onnx_cdist(
            X, Y, metric="euclidean", dtype=dtype, op_version=op_version)
    elif optim == 'cdist':
        dists = OnnxCDist(X, Y, metric="euclidean", op_version=op_version)
    else:
        raise ValueError("Unknown optimization '{}'.".format(optim))
    t_pi = py_make_float_array(pi, dtype=dtype)
    t_periodicity = py_make_float_array(periodicity, dtype)
    arg = OnnxMul(OnnxDiv(dists, t_periodicity, op_version=op_version),
                  t_pi, op_version=op_version)
    sin_of_arg = OnnxSin(arg, op_version=op_version)
    t_2 = py_make_float_array(2, dtype=dtype)
    t__2 = py_make_float_array(-2, dtype=dtype)
    t_length_scale = py_make_float_array(length_scale, dtype=dtype)
    K = OnnxExp(
        OnnxMul(
            OnnxPow(
                OnnxDiv(sin_of_arg, t_length_scale, op_version=op_version),
                t_2, op_version=op_version),
            t__2, op_version=op_version),
        op_version=op_version)
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_dot_product(X, Y, sigma_0=2.0, dtype=None, op_version=None,
                         **kwargs):
    """
    Implements the kernel DotProduct.
    :math:`k(x_i,x_j)=\\sigma_0^2+x_i\\cdot x_j`.
    """
    # It only works in two dimensions.
    t_sigma_0 = py_make_float_array(sigma_0 ** 2, dtype=dtype)
    if isinstance(Y, np.ndarray):
        tr = Y.T.astype(dtype)
    else:
        tr = OnnxTranspose(Y, perm=[1, 0], op_version=op_version)
    matm = OnnxMatMul(X, tr, op_version=op_version)
    K = OnnxAdd(matm, t_sigma_0, op_version=op_version)
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_rational_quadratic(X, Y, length_scale=1.0, alpha=2.0,
                                dtype=None, optim=None, op_version=None,
                                **kwargs):
    """
    Implements the kernel RationalQuadratic.
    :math:`k(x_i,x_j)=(1 + d(x_i, x_j)^2 / (2*\\alpha * l^2))^{-\\alpha}`.
    """
    if optim is None:
        dists = onnx_cdist(X, Y, dtype=dtype, metric="sqeuclidean",
                           op_version=op_version)
    elif optim == 'cdist':
        dists = OnnxCDist(X, Y, metric="sqeuclidean", op_version=op_version)
    else:
        raise ValueError("Unknown optimization '{}'.".format(optim))
    cst = length_scale ** 2 * alpha * 2
    t_cst = py_make_float_array(cst, dtype=dtype)
    tmp = OnnxDiv(dists, t_cst, op_version=op_version)
    t_one = py_make_float_array(1, dtype=dtype)
    base = OnnxAdd(tmp, t_one, op_version=op_version)
    t_alpha = py_make_float_array(-alpha, dtype=dtype)
    K = OnnxPow(base, t_alpha, op_version=op_version)
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_pairwise_kernel(X, Y, metric=None,
                             dtype=None, op_version=None,
                             gamma=None, gamma_bounds=None,
                             degree=None, coef0=None,
                             optim=None, **kwargs):
    """
    Implements the kernel PairwiseKernel.

    * cosine: :math:`k(x_i,x_j)=\\frac{<x_i, x_j>}
        {\\parallel x_i \\parallel \\parallel x_j \\parallel}`

    See `KERNEL_PARAMS
    <https://github.com/scikit-learn/scikit-learn/blob/
    95119c13af77c76e150b753485c662b7c52a41a2/sklearn/
    metrics/pairwise.py#L1862>`_.
    """
    if metric == 'cosine':
        if isinstance(Y, np.ndarray):
            ny = np.sqrt(np.sum(Y ** 2, axis=1, keepdims=True))
            norm_y = Y / ny
            norm_try = norm_y.T.astype(dtype)
        else:
            ny = OnnxReduceL2_typed(dtype, Y, axes=[1], op_version=op_version)
            norm_y = OnnxDiv(Y, ny, op_version=op_version)
            norm_try = OnnxTranspose(norm_y, perm=[1, 0],
                                     op_version=op_version)

        nx = OnnxReduceL2_typed(dtype, X, axes=[1], op_version=op_version)
        norm_x = OnnxDiv(X, nx, op_version=op_version)
        K = OnnxMatMul(norm_x, norm_try, op_version=op_version)
        return OnnxIdentity(K, op_version=op_version, **kwargs)
    raise NotImplementedError("Metric %r is not implemented." % metric)


def convert_kernel(kernel, X, output_names=None,
                   x_train=None, dtype=None, optim=None,
                   op_version=None):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if isinstance(kernel, Sum):
        clop = OnnxAdd
    elif isinstance(kernel, Product):
        clop = OnnxMul
    else:
        clop = None
    if clop is not None:
        return clop(
            convert_kernel(kernel.k1, X, x_train=x_train, dtype=dtype,
                           optim=optim, op_version=op_version),
            convert_kernel(kernel.k2, X, x_train=x_train, dtype=dtype,
                           optim=optim, op_version=op_version),
            output_names=output_names, op_version=op_version)

    if isinstance(kernel, ConstantKernel):
        # X and x_train should have the same number of features.
        if x_train is None:
            onnx_zeros_x = _zero_vector_of_size(
                X, keepdims=1, dtype=dtype, op_version=op_version)
            onnx_zeros_y = onnx_zeros_x
        else:
            onnx_zeros_x = _zero_vector_of_size(
                X, keepdims=1, dtype=dtype, op_version=op_version)
            onnx_zeros_y = _zero_vector_of_size(
                x_train, keepdims=1, dtype=dtype, op_version=op_version)

        tr = OnnxTranspose(onnx_zeros_y, perm=[1, 0], op_version=op_version)
        mat = OnnxMatMul(onnx_zeros_x, tr, op_version=op_version)
        return OnnxAdd(mat,
                       np.array([kernel.constant_value],
                                dtype=dtype),
                       output_names=output_names,
                       op_version=op_version)

    if isinstance(kernel, RBF):
        # length_scale = np.squeeze(length_scale).astype(float)
        zeroh = _zero_vector_of_size(X, axis=1, keepdims=0, dtype=dtype,
                                     op_version=op_version)
        zerov = _zero_vector_of_size(X, axis=0, keepdims=1, dtype=dtype,
                                     op_version=op_version)

        if (isinstance(kernel.length_scale, np.ndarray) and
                len(kernel.length_scale) > 0):
            const = kernel.length_scale.astype(dtype)
        else:
            tensor_value = py_make_float_array(
                kernel.length_scale, dtype=dtype, as_tensor=True)
            const = OnnxConstantOfShape(
                OnnxShape(zeroh, op_version=op_version),
                value=tensor_value, op_version=op_version)
        X_scaled = OnnxDiv(X, const, op_version=op_version)
        if x_train is None:
            dist = onnx_squareform_pdist(
                X_scaled, metric='sqeuclidean', dtype=dtype,
                op_version=op_version)
        else:
            x_train_scaled = OnnxDiv(x_train, const, op_version=op_version)
            if optim is None:
                dist = onnx_cdist(X_scaled, x_train_scaled,
                                  metric='sqeuclidean',
                                  dtype=dtype, op_version=op_version)
            elif optim == 'cdist':
                dist = OnnxCDist(X_scaled, x_train_scaled,
                                 metric='sqeuclidean',
                                 op_version=op_version)
            else:
                raise ValueError("Unknown optimization '{}'.".format(optim))

        tensor_value = py_make_float_array(-0.5, dtype=dtype, as_tensor=True)
        cst5 = OnnxConstantOfShape(
            OnnxShape(zerov, op_version=op_version),
            value=tensor_value, op_version=op_version)

        # K = np.exp(-.5 * dists)
        exp = OnnxExp(OnnxMul(dist, cst5, op_version=op_version),
                      output_names=output_names, op_version=op_version)

        # This should not be needed.
        # K = squareform(K)
        # np.fill_diagonal(K, 1)
        return exp

    if isinstance(kernel, ExpSineSquared):
        if not isinstance(kernel.length_scale, (float, int)):
            raise NotImplementedError(
                "length_scale should be float not {}.".format(
                    type(kernel.length_scale)))

        return _convert_exp_sine_squared(
            X, Y=X if x_train is None else x_train,
            length_scale=kernel.length_scale,
            periodicity=kernel.periodicity, dtype=dtype,
            output_names=output_names, optim=optim,
            op_version=op_version)

    if isinstance(kernel, DotProduct):
        if not isinstance(kernel.sigma_0, (float, int)):
            raise NotImplementedError(
                "sigma_0 should be float not {}.".format(
                    type(kernel.sigma_0)))

        if x_train is None:
            return _convert_dot_product(X, X, sigma_0=kernel.sigma_0,
                                        dtype=dtype,
                                        output_names=output_names,
                                        op_version=op_version)
        else:
            if len(x_train.shape) != 2:
                raise NotImplementedError(
                    "Only DotProduct for two dimension train set is "
                    "implemented.")
            return _convert_dot_product(
                X, x_train, sigma_0=kernel.sigma_0,
                dtype=dtype, output_names=output_names,
                op_version=op_version)

    if isinstance(kernel, RationalQuadratic):
        if x_train is None:
            return _convert_rational_quadratic(
                X, X, length_scale=kernel.length_scale,
                dtype=dtype, alpha=kernel.alpha,
                output_names=output_names,
                optim=optim, op_version=op_version)
        else:
            return _convert_rational_quadratic(
                X, x_train, length_scale=kernel.length_scale,
                dtype=dtype, alpha=kernel.alpha,
                output_names=output_names,
                optim=optim, op_version=op_version)

    if isinstance(kernel, PairwiseKernel):
        if x_train is None:
            return _convert_pairwise_kernel(
                X, X, metric=kernel.metric,
                dtype=dtype, output_names=output_names,
                optim=optim, op_version=op_version)
        else:
            return _convert_pairwise_kernel(
                X, x_train, metric=kernel.metric,
                dtype=dtype, output_names=output_names,
                optim=optim, op_version=op_version)

    raise RuntimeError("Unable to convert __call__ method for "
                       "class {}.".format(type(kernel)))


def _zero_vector_of_size(X, output_names=None, axis=0,
                         keepdims=None, dtype=None, op_version=None):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if keepdims is None:
        raise ValueError("Default for keepdims is not allowed.")
    if dtype == np.float32:
        res = OnnxReduceSumApi11(
            OnnxConstantOfShape(
                OnnxShape(X, op_version=op_version),
                op_version=op_version),
            axes=[1-axis], keepdims=keepdims,
            output_names=output_names, op_version=op_version)
    elif dtype in (np.float64, np.int32, np.int64):
        res = OnnxReduceSumApi11(
            OnnxConstantOfShape(
                OnnxShape(X, op_version=op_version), value=py_make_float_array(
                    0, dtype=dtype, as_tensor=True), op_version=op_version),
            axes=[1-axis], keepdims=keepdims,
            output_names=output_names, op_version=op_version)
    else:
        raise NotImplementedError(
            "Unable to create zero vector of type {}".format(dtype))
    return res



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType
)
from ..common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types
)


def calculate_gaussian_mixture_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=[2, 3])
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType])

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Input must be a [N, C]-tensor')

    op = operator.raw_operator
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = Int64TensorType([N, 1])
    operator.outputs[1].type.shape = [N, op.n_components]
    if len(operator.outputs) > 2:
        operator.outputs[2].type.shape = [N, 1]


register_shape_calculator('SklearnGaussianMixture',
                          calculate_gaussian_mixture_output_shapes)
register_shape_calculator('SklearnBayesianGaussianMixture',
                          calculate_gaussian_mixture_output_shapes)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType, StringTensorType)
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_imputer_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C_1], ..., [N, C_n] ---> [N, C_1 + ... + C_n]

    It's possible to receive multiple inputs so we need to concatenate
    them along C-axis. The produced tensor's shape is used as the
    output shape.
    """
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType,
            StringTensorType])
    if type(operator.inputs[0].type) != type(operator.outputs[0].type):  # noqa
        raise RuntimeError(
            "Inputs and outputs should have the same type "
            "%r != %r." % (
                type(operator.inputs[0].type),
                type(operator.outputs[0].type)))

    N = operator.inputs[0].get_first_dimension()
    C = 0
    for variable in operator.inputs:
        if variable.type.shape[1] is not None:
            C += variable.type.shape[1]
        else:
            C = None
            break

    operator.outputs[0].type.shape = [N, C]


register_shape_calculator('SklearnImputer',
                          calculate_sklearn_imputer_output_shapes)
register_shape_calculator('SklearnSimpleImputer',
                          calculate_sklearn_imputer_output_shapes)
register_shape_calculator('SklearnBinarizer',
                          calculate_sklearn_imputer_output_shapes)
